package org.processmining.performancespectrumintegration.algorithms.miner;

import com.google.common.collect.Multimap;
import com.google.common.collect.MultimapBuilder;
import com.google.common.collect.Sets;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.performancespectrumintegration.algorithms.petrinet.AdjPlaces;
import org.processmining.performancespectrumintegration.algorithms.petrinet.ExtendedMarking;

import java.util.*;

import static org.processmining.performancespectrumintegration.algorithms.miner.AlignmentSemanticsBuilder.ExecutionType.*;

public class AlignmentSemanticsBuilder<T extends SegmentTemplateBuilder> {

    private final Map<Transition, AdjPlaces> affectedPlaces;
    private final ExtendedMarking<TransitionExecution> current;
    private final Multimap<TransitionExecution, TransitionExecution> dependencyMap;
    private final AlignmentSemantics alignmentSemantics;

    public AlignmentSemanticsBuilder(Marking initialMarking, Map<Transition, AdjPlaces> adjPlacesMap) {
        this.current = new ExtendedMarking<>();
        addInitialEnablers(initialMarking);
        this.affectedPlaces = adjPlacesMap;
        this.dependencyMap = MultimapBuilder.hashKeys().hashSetValues().build();
        alignmentSemantics = new AlignmentSemantics();
    }

    private void addInitialEnablers(Marking initialMarking) {
        for (Place place : initialMarking) {
            current.addTo(place, new TransitionExecution(-1, Artificial));
        }
    }

    private SortedSet<TransitionExecution> getEnablingTransitionExecutions(Transition t) {
        return current.peekFrom(affectedPlaces.get(t).getPreset());
    }

    private TransitionExecution getFinalEnablingTransitionExecution(Transition t) {
        SortedSet<TransitionExecution> texs = getEnablingTransitionExecutions(t);

        if (texs.isEmpty()) return new TransitionExecution(-1, Artificial);

        Set<TransitionExecution> dependents = new HashSet<>();

        TransitionExecution finalTeX = new TransitionExecution(-2, Artificial);
        boolean containsModelMove = false;
        for (TransitionExecution enabler : texs) {
            if (enabler.eventIndex > finalTeX.eventIndex) finalTeX = enabler;
            if (enabler.executionType == Artificial) dependents.add(enabler);
            containsModelMove = containsModelMove || (enabler.executionType == Model);
        }

        if (containsModelMove) return new TransitionExecution(-1, Model);

        dependents.remove(finalTeX);

        dependencyMap.putAll(finalTeX, dependents);
        return finalTeX;
    }

    private void executeTransition(Transition t, TransitionExecution tex) {
        AdjPlaces triple = affectedPlaces.get(t);
        current.removeFrom(triple.getPreset());
        current.addTo(triple.getPostset(), tex);
        alignmentSemantics.add(t, tex);
    }

    public void syncMove(Transition t, int eventIndex) {
        SortedSet<TransitionExecution> texs = getEnablingTransitionExecutions(t); // those that have "artificial" enablers and can be updated
        TransitionExecution tex = new TransitionExecution(eventIndex, Sync);
        dependencyMap.putAll(tex, texs.headSet(TransitionExecution.ARTIFICIAL_ZERO));
        updateTransitively(tex, eventIndex);
        executeTransition(t, tex);
    }

    private void updateTransitively(TransitionExecution tex, int eventIndex) {
        LinkedList<TransitionExecution> toDo = new LinkedList<>(); // stack -> depth first, queue -> breadth
        Set<TransitionExecution> done = Sets.newHashSet();

        if (dependencyMap.containsKey(tex)) {
            toDo.addAll(dependencyMap.get(tex));
            done.addAll(dependencyMap.get(tex));
            dependencyMap.removeAll(tex);
        }

        while (!toDo.isEmpty()) {
            TransitionExecution te = toDo.removeFirst();

            // TODO DEBUG further debugging would be interesting
            // System.out.println("set " + te.executionType + " " + te.eventIndex + " to sync " + eventIndex);

            te.executionType = Sync;
            te.eventIndex = eventIndex;
            if (dependencyMap.containsKey(te)) {
                for (TransitionExecution te2 : dependencyMap.get(te)) {
                    if (!done.contains(te2)) {
                        toDo.addLast(te2);
                        done.add(te2);
                    }
                }
                // dependencyMap.values().removeAll(set);
                dependencyMap.removeAll(te);
            }
        }
    }

    public void modelMove(Transition t) {
        executeTransition(t, new TransitionExecution(-1, Model));
    }

    public void invMove(Transition t) {
        TransitionExecution tex = getFinalEnablingTransitionExecution(t);
        if (tex.executionType == Sync) updateTransitively(tex, tex.eventIndex);
        executeTransition(t, tex);
    }

    public AlignmentSemantics getAlignmentSemantics() {
        return alignmentSemantics;
    }

    enum ExecutionType {
        Artificial, Model, Sync
    }

}
