package org.processmining.performancespectrumintegration.algorithms.miner;

import com.google.common.collect.*;
import org.apache.commons.math3.stat.descriptive.rank.Percentile;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.jfree.data.category.CategoryDataset;
import org.jfree.data.category.DefaultCategoryDataset;
import org.jfree.data.statistics.HistogramDataset;
import org.jfree.data.statistics.HistogramType;
import org.joda.time.DateTime;
import org.joda.time.Duration;
import org.joda.time.Instant;
import org.joda.time.Period;
import org.processmining.log.utils.XUtils;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.performancespectrumintegration.algorithms.segments.Segment;
import org.processmining.performancespectrumintegration.models.settings.*;
import org.processmining.performancespectrumintegration.util.ToStrings;
import org.processmining.performancespectrumintegration.util.Tuple2;
import org.processmining.plugins.petrinet.replayresult.PNRepResult;
import org.processmining.plugins.petrinet.replayresult.StepTypes;
import org.processmining.plugins.replayer.replayresult.SyncReplayResult;

import java.util.*;
import java.util.function.BiConsumer;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.stream.Collectors;


@SuppressWarnings("Duplicates")
public class LogAlignmentInfoExtractor {

    private final XLog log;
    private final PNRepResult replayResults;

    public LogAlignmentInfoExtractor(XLog log, PNRepResult replayResults) {
        this.log = log;
        this.replayResults = replayResults;
    }

    private static void applyToCaseDurationMap(BiConsumer<Integer, Long> consumer, Map<Integer, Long> caseDurationMap) {
        for (Map.Entry<Integer, Long> entry : caseDurationMap.entrySet()) {
            consumer.accept(entry.getKey(), entry.getValue());
        }
    }

    public void runOverAlignments(Function<Integer, BiConsumer<DateTime, DateTime>> action) {
        runOverFilteredAlignments(i -> true, action);
    }

    public void runOverFilteredAlignments(CaseIdFilter selectedCaseIds, Function<Integer, BiConsumer<DateTime, DateTime>> action) {
        for (SyncReplayResult replayResult : replayResults) {
            List<StepTypes> types = replayResult.getStepTypes();

            int eventCounter = 0;
            int startIndex = -1, endIndex = -1;
            for (StepTypes type : types) {
                if (type == StepTypes.L) eventCounter++;
                else if (type == StepTypes.LMGOOD) {
                    if (startIndex < 0)
                        startIndex = eventCounter;
                    endIndex = eventCounter;
                    eventCounter++;
                }
            }

            if (startIndex >= 0) { // otherwise there were no log or sync moves at all
                for (Integer index : replayResult.getTraceIndex()) {
                    if (selectedCaseIds.isIncluded(index)) {
                        XTrace trace = log.get(index);
                        DateTime startTime = new DateTime(XUtils.getTimestamp(trace.get(startIndex)));
                        DateTime endTime = new DateTime(XUtils.getTimestamp(trace.get(endIndex)));
                        action.apply(index).accept(startTime, endTime);
                    }
                }
            }
        }
    }

    public CaseIdClassification getDurationThresholdClassifier(CaseIdFilter selectedCaseIds, Duration splitDuration) {
        return getDurationThresholdClassifier(selectedCaseIds, splitDuration, null);
    }

    public CaseIdClassification getDurationThresholdClassifier(CaseIdFilter selectedCaseIds, Duration splitDuration, Map<Integer, Long> caseDurationMap) {
        SortedMap<Integer, String> legend = Maps.newTreeMap();
        String print = ToStrings.print(splitDuration);
        legend.put(0, "faster than " + print + " (inclusive)");
        legend.put(1, "slower than " + print);

        Map<Integer, Integer> resultMap = Maps.newHashMap();

        long millis = splitDuration.getMillis();
        BiConsumer<Integer, Long> consumer = (i, d) -> {
            resultMap.put(i, d <= millis ? 0 : 1);
        };

        if (caseDurationMap == null)
            runOverFilteredAlignments(selectedCaseIds, i -> (DateTime startTime, DateTime endTime) -> {
                long d = endTime.getMillis() - startTime.getMillis();
                consumer.accept(i, d);
            });
        else
            applyToCaseDurationMap(consumer, caseDurationMap);

        return new MapBasedClassification(resultMap, legend);
    }

    public CaseIdClassification getBinaryDurationClassifier(CaseIdFilter caseIdFilter, double topPercentage) {
        Map<Integer, Long> caseDurationMap = Maps.newHashMap();

        List<Double> temp = new ArrayList<>();
        runOverFilteredAlignments(caseIdFilter, i -> (d1, d2) -> {
            long duration = d2.getMillis() - d1.getMillis();
            caseDurationMap.put(i, duration);
            temp.add((double) duration);
        });

        Percentile percentile = new Percentile();
        percentile.setData(temp.stream().mapToDouble(Double::doubleValue).toArray());

        Duration threshold = new Duration((long) percentile.evaluate(100 * topPercentage));

        return getDurationThresholdClassifier(caseIdFilter, threshold, caseDurationMap);
    }

    public CaseIdClassification getDurationClassesClassifier(CaseIdFilter caseIdFilter, int groups) {
        Map<Integer, Long> caseDurationMap = Maps.newHashMap();
        SortedMap<Integer, String> legend = Maps.newTreeMap();

        List<Double> temp = new ArrayList<>();
        runOverFilteredAlignments(caseIdFilter, i -> (d1, d2) -> {
            long duration = d2.getMillis() - d1.getMillis();
            caseDurationMap.put(i, duration);
            temp.add((double) duration);
        });

        Percentile percentile = new Percentile();
        percentile.setData(temp.stream().mapToDouble(Double::doubleValue).toArray());

        double[] percentiles = new double[groups];
        for (int i = 0; i < groups; i++) {
            percentiles[i] = percentile.evaluate((double) (i + 1) / groups * 100);
            Period period = new Period(new Instant(0), new Instant((long) percentiles[i]));
            String print = ToStrings.print(period);
            legend.put(i, String.format("%.2f-%.2f%% percentiles of case duration (up to %s)", 100 * (double) i / groups, 100 * (double) (i + 1) / groups, print));
        }

        Map<Integer, Integer> resultMap = Maps.newHashMap();

        for (Map.Entry<Integer, Long> entry : caseDurationMap.entrySet()) {
            long duration = entry.getValue();
            int i = 0;
            while (i < percentiles.length - 1 && duration > percentiles[i]) {
                i++;
            }
            resultMap.put(entry.getKey(), i);
        }

        return new MapBasedClassification(resultMap, legend);
    }

    public CaseIdClassification getBinaryVariantClassifier(CaseIdFilter caseIdFilter, double topPercentage) {
        Multimap<Integer, Set<Integer>> variantMap = MultimapBuilder.<Integer>treeKeys(Comparator.reverseOrder()).arrayListValues().build();

        int totalSize = 0;
        for (SyncReplayResult replayResult : replayResults) {
            Set<Integer> intersection = caseIdFilter.applyFilterToImmutable(replayResult.getTraceIndex());
            int size = intersection.size();
            totalSize += size;
            variantMap.put(size, intersection);
        }

        Map<Integer, Integer> result = Maps.newHashMap();

        int currentSize = 0;
        for (Map.Entry<Integer, Collection<Set<Integer>>> entry : variantMap.asMap().entrySet()) {
            Integer size = entry.getKey();

            for (Set<Integer> caseIds : entry.getValue()) {
                if (currentSize / (double) totalSize < topPercentage) {
                    caseIds.forEach(i -> result.put(i, 0));
                    currentSize += size;
                } else {
                    caseIds.forEach(i -> result.put(i, 1));
                }
            }

        }

        double actualPercentage = currentSize / (double) totalSize;

        SortedMap<Integer, String> legend = Maps.newTreeMap();
        legend.put(0, String.format("top %.2f%% variants", 100 * actualPercentage));
        legend.put(1, String.format("bottom %.2f%% variants", 100 * (1 - actualPercentage)));

        return new MapBasedClassification(result, legend);
    }

    public SetBasedFilter getDurationPercentileFilter(double topPercentage, boolean invert) {
        Map<Integer, Long> caseDurationMap = Maps.newHashMap();

        List<Double> temp = new ArrayList<>();
        runOverAlignments(i -> (d1, d2) -> {
            long duration = d2.getMillis() - d1.getMillis();
            caseDurationMap.put(i, duration);
            temp.add((double) duration);
        });

        Percentile percentile = new Percentile();
        percentile.setData(temp.stream().mapToDouble(i -> i).toArray());
        double highestDuration = percentile.evaluate((invert ? (1 - topPercentage) : topPercentage) * 100);

        return getDurationThresholdFilter(new Duration((long) highestDuration), invert, caseDurationMap);
    }

    public SetBasedFilter getDurationThresholdFilter(Duration thresholdDuration, boolean invert) {
        return getDurationThresholdFilter(thresholdDuration, invert, null);
    }

    public SetBasedFilter getDurationThresholdFilter(Duration thresholdDuration, boolean invert, Map<Integer, Long> caseDurationMap) {
        Set<Integer> result = Sets.newHashSet();

        long millis = thresholdDuration.getMillis();
        Predicate<Long> pred;
        if (invert) pred = d -> d >= millis;
        else pred = d -> d <= millis;

        BiConsumer<Integer, Long> consumer = (i, d) -> {
            if (pred.test(d)) result.add(i);
        };

        if (caseDurationMap == null)
            runOverAlignments(i -> (d1, d2) -> {
                long d = d2.getMillis() - d1.getMillis();
                consumer.accept(i, d);
            });
        else
            applyToCaseDurationMap(consumer, caseDurationMap);

        return new SetBasedFilter(result);
    }

    public SetBasedFilter getVariantFilter(double percentage, boolean invert) {
        Comparator<Integer> comp;
        if (invert) comp = Comparator.naturalOrder(); // small to big
        else comp = Comparator.reverseOrder(); // big to small

        Multimap<Integer, Set<Integer>> variantMap = MultimapBuilder.<Integer>treeKeys(comp).arrayListValues().build();

        int totalSize = 0;
        for (SyncReplayResult replayResult : replayResults) {
            SortedSet<Integer> traceIndex = replayResult.getTraceIndex();
            int size = traceIndex.size();
            variantMap.put(size, traceIndex);
            totalSize += size;
        }

        Set<Integer> result = Sets.newHashSet();

        int currentSize = 0;
        for (Map.Entry<Integer, Collection<Set<Integer>>> entry : variantMap.asMap().entrySet()) {
            Integer size = entry.getKey();

            for (Set<Integer> caseIds : entry.getValue()) {
                if (currentSize / (double) totalSize < percentage) {
                    result.addAll(caseIds);
                }
                currentSize += size;
            }

        }

        return new SetBasedFilter(result);
    }

    public Tuple2<CategoryDataset, List<String>> getVariantDistribution(CaseIdFilter caseIdFilter) {
        SortedSet<Tuple2<Integer, String>> sizes = Sets.newTreeSet(Comparator.comparing((Tuple2<Integer, String> t) -> t._1).reversed());
        double totalSize = 0;
        for (SyncReplayResult replayResult : replayResults) {
            int size = (int) caseIdFilter.countIntersection(replayResult.getTraceIndex());
            totalSize += size;
            sizes.add(new Tuple2<>(size, ToStrings.print(replayResult)));
        }
        List<String> variantLabels = new ArrayList<>();
        DefaultCategoryDataset categoryDataset = new DefaultCategoryDataset();
        int i = 0;
        for (Tuple2<Integer, String> t : sizes) {
            categoryDataset.addValue((Number) (t._1 / totalSize), i++, "");
            variantLabels.add(t._2);
        }

        return new Tuple2<>(categoryDataset, variantLabels);
    }

    public HistogramDataset getDurationHistogramDataset(CaseIdFilter caseIdFilter) {
        HistogramDataset histogramDataset = new HistogramDataset();

        ArrayList<Double> temp = new ArrayList<>();
        runOverFilteredAlignments(caseIdFilter, i -> (d1, d2) -> {
            double duration = d2.getMillis() - d1.getMillis();
            temp.add(duration);
        });
        histogramDataset.setType(HistogramType.RELATIVE_FREQUENCY);

        histogramDataset.addSeries("Case Durations", temp.stream().mapToDouble(i -> i).toArray(), 10);
        return histogramDataset;
    }

    public Tuple2<DateTime, DateTime> getProcessDuration(CaseIdFilter selectedCaseIds) {
        Tuple2<DateTime, DateTime> t;
        t = new Tuple2<>();

        runOverFilteredAlignments(selectedCaseIds, i -> (d1, d2) -> {
            if (t._1 == null || d1.isBefore(t._1)) t._1 = d1;
            if (t._2 == null || d2.isAfter(t._2)) t._2 = d2;
        });

        return t;
    }

    public LinkedList<Segment> getSegmentOrder(Set<Segment> segments, Filter<Integer> selectedCases) {
        Table<Segment, Segment, Integer> df = HashBasedTable.create();

        int maxDf = 0;

        Set<Segment> endedSegments = Sets.newHashSet();

        Segment fakeStart = new Segment() {
            @Override
            public ImmutableSet<Transition> getFrom() {
                return null;
            }

            @Override
            public ImmutableSet<Transition> getTo() {
                return null;
            }

            @Override
            public String toString() {
                return "start";
            }
        };

        for (SyncReplayResult rR : replayResults) {
            int size = (int) selectedCases.countIntersection(rR.getTraceIndex());

            if (size == 0) continue;

            endedSegments.clear();
            endedSegments.add(fakeStart);

            int k = 0;
            for (StepTypes stepType : rR.getStepTypes()) {
                switch (stepType) {
                    case LMGOOD:
                    case MINVI:
                    case MREAL:
                        Transition t = ((Transition) rR.getNodeInstance().get(k));
                        for (Segment segment : segments) {
                            if (segment.getTo().contains(t)) {
                                endedSegments.add(segment);
                            }
                        }
                        boolean newDf = false;
                        for (Segment segment : segments) {
                            if (segment.getFrom().contains(t)) {
                                for (Segment endedSegment : endedSegments) {
                                    Integer integer = df.get(endedSegment, segment);
                                    if (integer == null) {
                                        integer = 0;
                                    }
                                    df.put(endedSegment, segment, integer + size);
                                    maxDf = Math.max(maxDf, integer + size);
                                    newDf = true;
                                }
                            }
                        }
                        if (newDf) endedSegments.clear();
                    default:
                        k++;
                        break;
                }
            }
        }


        int max = maxDf;
        df.cellSet().removeIf(cell -> (double) cell.getValue() / max < .1);

        // should also take care of self loops
        for (Segment r : segments) {
            for (Segment c : segments) {
                if (df.contains(r, c) && df.contains(c, r)) {
                    if (df.get(r, c) > df.get(c, r))
                        df.remove(c, r);
                    else
                        df.remove(r, c);
                }
            }
        }

        LinkedList<Segment> ordered = new LinkedList<>();
        LinkedList<Segment> unseen = new LinkedList<>(segments);
        LinkedList<Segment> toDo = new LinkedList<>();

        unseen.addFirst(fakeStart);

        while (!unseen.isEmpty()) {
            toDo.add(unseen.getFirst());

            while (!toDo.isEmpty()) {
                Segment first = toDo.removeFirst();

                if (first != fakeStart) {
                    ordered.add(first);
                }
                unseen.remove(first);

                for (Map.Entry<Segment, Integer> entry : df.row(first).entrySet()) {
                    Segment key = entry.getKey();
                    if (unseen.contains(key))
                        toDo.addFirst(key); // stack -> depth first
                }
            }
        }

        return ordered;
    }

    public CaseIdFilter buildNoFilter() {
        return i -> true;
    }

    public MapBasedClassification buildNoClassifier(Filter<Integer> caseIdFilter) {
        Map<Integer, Integer> collect = replayResults.stream().flatMap(rR -> rR.getTraceIndex().stream()).filter(caseIdFilter::isIncluded).collect(Collectors.toMap(i -> i, i -> 0));
        SortedMap<Integer, String> legend = Maps.newTreeMap();
        legend.put(0, "default");
        return new MapBasedClassification(collect, legend);
    }

}
