package org.processmining.performancespectrumintegration.algorithms.miner;

import com.google.common.collect.*;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.joda.time.DateTime;
import org.processmining.acceptingpetrinet.models.AcceptingPetriNet;
import org.processmining.log.utils.XUtils;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.performancespectrumintegration.algorithms.petrinet.AdjPlaces;
import org.processmining.performancespectrumintegration.algorithms.petrinet.GraphTools;
import org.processmining.performancespectrumintegration.algorithms.segments.OrderedSegmentMeasurement;
import org.processmining.performancespectrumintegration.algorithms.segments.Segment;
import org.processmining.performancespectrumintegration.algorithms.segments.SegmentOccurrence;
import org.processmining.performancespectrumintegration.models.settings.Filter;
import org.processmining.plugins.petrinet.replayresult.PNRepResult;
import org.processmining.plugins.petrinet.replayresult.StepTypes;
import org.processmining.plugins.replayer.replayresult.SyncReplayResult;

import java.util.*;
import java.util.stream.Collectors;

public class SegmentDataMiner {

    private static final boolean UPDATE_FILTER_DELTA = false;
    private static final int MAX_TEMPLATE_CACHE_SIZE = 30;
    private static final int MAX_SEGMENTS_CACHE_SIZE = 10;

    private final XLog log;
    private final Petrinet net;
    private final List<SyncReplayResult> replayResults;
    private final List<AlignmentSemantics> alignmentSemanticsList;
    private final Marking initialMarking;
    private final Map<Transition, AdjPlaces> adjPlacesMap;
    private final SegmentMiningParameters lastParameters;
    private List<Multimap<Segment, SegmentOccurrenceTemplate>> cachedTemplates;
    private SegmentData cachedSegmentData;
    private Set<Segment> cachedSegmentTemplates;
    private Set<Segment> cachedSegments;
    private Boolean areTemplatesDirty;

    public SegmentDataMiner(XLog log, AcceptingPetriNet acceptingPetriNet, PNRepResult replayResult) {
        this.log = log;
        this.net = acceptingPetriNet.getNet();
        this.replayResults = new ArrayList<>(replayResult);
        this.initialMarking = acceptingPetriNet.getInitialMarking();
        this.adjPlacesMap = GraphTools.getAdjPlacesMap(net);
        this.lastParameters = new SegmentMiningParameters(null, null);
        this.alignmentSemanticsList = buildAlignmentSemantics();
        this.cachedSegmentTemplates = Sets.newHashSet();
        this.cachedSegments = Sets.newHashSet();
    }

    private List<AlignmentSemantics> buildAlignmentSemantics() {
        List<AlignmentSemantics> collect = replayResults.stream().map(rR -> {
            List<Object> nodeInstance = rR.getNodeInstance();
            List<StepTypes> stepTypes = rR.getStepTypes();

            AlignmentSemanticsBuilder<SegmentTemplateBuilder> semanticsBuilder = new AlignmentSemanticsBuilder<>(initialMarking, adjPlacesMap);

            int eventIndex = 0;

            for (int alignmentIndex = 0; alignmentIndex < nodeInstance.size(); alignmentIndex++) {
                Object o = nodeInstance.get(alignmentIndex);
                StepTypes stepType = stepTypes.get(alignmentIndex);

                switch (stepType) {
                    case LMGOOD:
                        semanticsBuilder.syncMove((Transition) o, eventIndex);
                        eventIndex++;
                        break;
                    case MREAL:
                        semanticsBuilder.modelMove((Transition) o);
                        break;
                    case MINVI:
                        semanticsBuilder.invMove((Transition) o);
                        break;
                    case L:
                        eventIndex++;
                        break;
                    default:
                        System.err.println("unexpected steptype in alignment: " + stepType);
                }
            }
            return semanticsBuilder.getAlignmentSemantics();
        }).collect(Collectors.toList());

        // TODO DEBUG
        // System.out.println(collect.stream().map(Object::toString).collect(Collectors.joining("\n--------------\n")));

        return collect;
    }

    // this requires that updateTemplates has been called beforehand
    public boolean updateCachedData(final Set<Segment> selectedSegments, final Filter<Integer> filter) {
        if (filter == null) return false;

        synchronized (lastParameters) {
            if (lastParameters.getFilter() == null) {
                cachedSegmentData = getRawDataSequential(selectedSegments, filter);
                lastParameters.setFilter(filter);
                cachedSegments = Sets.newHashSet(selectedSegments);
            } else if (filter == lastParameters.getFilter() && cachedSegments.containsAll(selectedSegments)) {
                return false;
            } else if (filter == lastParameters.getFilter()) {
                // the same filter as before, but not all rawData is cached

                ImmutableSet<Segment> newSegments = Sets.difference(selectedSegments, cachedSegments).immutableCopy();

                if (cachedSegments.size() + newSegments.size() > MAX_SEGMENTS_CACHE_SIZE) {
                    cachedSegmentData = getRawDataSequential(selectedSegments, filter);
                    cachedSegments = Sets.newHashSet(selectedSegments);
                } else {
                    SegmentData newRawSegmentData = getRawDataSequential(newSegments, filter);
                    cachedSegmentData.merge(newRawSegmentData);
                    cachedSegments.addAll(selectedSegments);
                }
            } else {

                if (UPDATE_FILTER_DELTA) {
                    Filter<Integer> differenceFilter = filter.difference(lastParameters.getFilter());
                    // TODO might actually be faster to just recompute for all cases

                    SegmentData newRawSegmentData = getRawDataSequential(selectedSegments, differenceFilter);
                    newRawSegmentData.merge(cachedSegmentData);
                    newRawSegmentData.restrictCases(filter);
                    cachedSegmentData = newRawSegmentData;
                } else
                    cachedSegmentData = getRawDataSequential(selectedSegments, filter);

                cachedSegments = Sets.newHashSet(selectedSegments);
                lastParameters.setFilter(filter);
            }
        }

        return true;
    }

    public SegmentData getSegmentData() {
        return cachedSegmentData;
    }

    private SegmentData getRawDataSequential(Set<Segment> selectedSegments, Filter<Integer> selectedCaseIds) {
        SegmentData segmentData = new SegmentData();

        int k = 0;
        for (SyncReplayResult rR : replayResults) {
            Multimap<Segment, SegmentOccurrenceTemplate> templateMultimap = cachedTemplates.get(k++);

            for (Segment selectedSegment : selectedSegments) {
                Collection<SegmentOccurrenceTemplate> collection = templateMultimap.get(selectedSegment);

                OrderedSegmentMeasurement sm = new OrderedSegmentMeasurement();

                if (collection != null && !collection.isEmpty())
                    for (Integer caseId : rR.getTraceIndex()) {
                        if (selectedCaseIds.isIncluded(caseId)) {
                            XTrace trace = log.get(caseId);

                            sm.incCaseCounter();

                            for (SegmentOccurrenceTemplate template : collection) {
                                if (template.involvesModelMove()) {
                                    sm.addUncountedOccurrence();
                                } else {
                                    Integer i = template.getFromEventIndex();
                                    Integer j = template.getToEventIndex();

                                    XEvent fromEvent = trace.get(i);
                                    XEvent toEvent = trace.get(j);

                                    sm.addNormalOccurrence(new SegmentOccurrence(new DateTime(XUtils.getTimestamp(fromEvent)), new DateTime(XUtils.getTimestamp(toEvent)), caseId));
                                }
                            }

                        }
                    }

                segmentData.add(selectedSegment, sm);

            }

        }

        return segmentData;
    }

    public boolean updateTemplates(final Set<Segment> selectedSegments) {
        if (selectedSegments == null) return false;

        boolean modifiedTemplates = false;
        synchronized (lastParameters) {
            ImmutableSet<Segment> immutableSet = ImmutableSet.copyOf(selectedSegments);

            if (lastParameters.getSelectedSegments() == null || cachedSegmentTemplates.size() > MAX_TEMPLATE_CACHE_SIZE) {
                lastParameters.setSelectedSegments(immutableSet);

                cachedTemplates = createSegmentOccurrenceTemplates(lastParameters.getSelectedSegments());
                cachedSegmentTemplates = Sets.newHashSet(immutableSet);

                modifiedTemplates = true;
            } else {

                ImmutableSet<Segment> segmentsToAdd = Sets.difference(immutableSet, cachedSegmentTemplates).immutableCopy();

                if (!segmentsToAdd.isEmpty()) {
                    List<Multimap<Segment, SegmentOccurrenceTemplate>> newTemplates = createSegmentOccurrenceTemplates(segmentsToAdd);
                    for (int i = 0; i < cachedTemplates.size(); i++) {
                        Multimap<Segment, SegmentOccurrenceTemplate> templateMultimap = cachedTemplates.get(i);
                        Multimap<Segment, SegmentOccurrenceTemplate> newTemplateMultimap = newTemplates.get(i);

                        templateMultimap.putAll(newTemplateMultimap); // will overwrite existing segments which should never happen
                    }

                    lastParameters.setSelectedSegments(immutableSet);
                    cachedSegmentTemplates.addAll(immutableSet);

                    modifiedTemplates = true;
                }
            }

            areTemplatesDirty = modifiedTemplates;
        }
        return modifiedTemplates;
    }

    private List<Multimap<Segment, SegmentOccurrenceTemplate>> createSegmentOccurrenceTemplates
            (Set<Segment> segmentSet) {
        return alignmentSemanticsList.stream().map(al -> createSegmentOccurrenceTemplatesForAlignment(segmentSet, al)).collect(Collectors.toList());
    }

    public Multimap<Segment, SegmentOccurrenceTemplate> createSegmentOccurrenceTemplatesForAlignment
            (Collection<Segment> segments, AlignmentSemantics alignmentSemantics) {
        Multimap<Segment, SegmentOccurrenceTemplate> segmentOccurrenceTemplates = MultimapBuilder.hashKeys().linkedListValues().build();

        if (segments.isEmpty()) return segmentOccurrenceTemplates;

        Map<Segment, SegmentTemplateBuilder> builders = Maps.newHashMapWithExpectedSize(segments.size());
        Multimap<Transition, SegmentTemplateBuilder> fromTriggers = MultimapBuilder.hashKeys().linkedListValues().build();
        Multimap<Transition, SegmentTemplateBuilder> toTriggers = MultimapBuilder.hashKeys().linkedListValues().build();
        for (Segment segment : segments) {
            SegmentTemplateBuilder stb = new SegmentTemplateBuilder();
            builders.put(segment, stb);
            for (Transition t : segment.getFrom()) {
                fromTriggers.put(t, stb);
            }
            for (Transition t : segment.getTo()) {
                toTriggers.put(t, stb);
            }
        }

        alignmentSemantics.applySegmentTemplateBuilders(fromTriggers, toTriggers);

        for (Map.Entry<Segment, SegmentTemplateBuilder> entry : builders.entrySet()) {
            segmentOccurrenceTemplates.putAll(entry.getKey(), entry.getValue().getOccurrenceTemplates());
        }

        return segmentOccurrenceTemplates;
    }

    private void restrictSegments() {
        cachedTemplates.forEach(mm -> mm.keySet().retainAll(lastParameters.getSelectedSegments()));
    }

}
