package org.processmining.performancespectrumintegration.algorithms.segments;

import com.google.common.collect.Sets;
import org.joda.time.DateTime;
import org.processmining.performancespectrumintegration.models.MeasurementReliability;
import org.processmining.performancespectrumintegration.models.settings.Filter;

import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.SortedSet;

public class OrderedSegmentMeasurement implements Iterable<SegmentOccurrence> {

    private final SortedSet<SegmentOccurrence> occurrences;
    private final MeasurementReliability measurementReliability;

    public OrderedSegmentMeasurement() {
        occurrences = Sets.newTreeSet();
        measurementReliability = new MeasurementReliability();
    }

    public MeasurementReliability getMeasurementReliability() {
        return measurementReliability;
    }

    public SortedSet<SegmentOccurrence> getOccurrences() {
        return occurrences;
    }

    public void addUncountedOccurrence() {
        measurementReliability.incUncountedOccurrences(1);
    }

    public void addNormalOccurrence(SegmentOccurrence segmentOccurrence) {
        assert segmentOccurrence != null;
        occurrences.add(segmentOccurrence);
        measurementReliability.incNormalOccurrences(1);
    }

    public SortedSet<SegmentOccurrence> subsetApproximation(DateTime from, DateTime to) {
        return occurrences.subSet(SegmentOccurrence.artificial(from), SegmentOccurrence.artificial(to));
    }

    public Iterable<SegmentOccurrence> subset(DateTime from, DateTime to) {
        return new Iterable<SegmentOccurrence>() {
            @Override
            public Iterator<SegmentOccurrence> iterator() {
                return new SubsetIterator(occurrences, from, to);
            }
        };
    }

    public OrderedSegmentMeasurement merge(OrderedSegmentMeasurement other) {
        assert other != null;
        occurrences.addAll(other.occurrences);
        measurementReliability.merge(other.measurementReliability);
        return this;
    }

    public void restrictCases(Filter<Integer> filter) {
        occurrences.removeIf(so -> filter.isNotIncluded(so.getCaseId()));
    }

    @Override
    public Iterator<SegmentOccurrence> iterator() {
        return occurrences.iterator();
    }

    public void incCaseCounter() {
        measurementReliability.incCaseCounter(1);
    }

    private static class SubsetIterator implements Iterator<SegmentOccurrence> {

        private final Iterator<SegmentOccurrence> it;
        private final DateTime from;
        private final DateTime to;

        private SegmentOccurrence next;

        SubsetIterator(SortedSet<SegmentOccurrence> occurrences, DateTime from, DateTime to) {
            this.from = from;
            this.to = to;
            it = occurrences.iterator();

            scroll();
        }

        private void scroll() {
            if (!it.hasNext()) {
                next = null;
                return;
            }

            do {
                next = it.next();
            }
            while (from.compareTo(next.getEndTime()) > 0 && it.hasNext());

            if (from.compareTo(next.getEndTime()) > 0 || to.compareTo(next.getStartTime()) < 0) next = null;
        }


        @Override
        public boolean hasNext() {
            return next != null;
        }

        @Override
        public SegmentOccurrence next() {
            if (next == null) throw new NoSuchElementException();
            SegmentOccurrence res = next;
            scroll();
            return res;
        }
    }
}
