package org.processmining.performancespectrumintegration.models;

import org.joda.time.DateTime;
import org.joda.time.Duration;
import org.joda.time.Instant;
import org.joda.time.Period;
import org.processmining.performancespectrumintegration.algorithms.segments.Segment;
import org.processmining.performancespectrumintegration.models.settings.AggregationIntervalSettings;

import java.util.Collection;

public abstract class AbstractTimeWindowData {

    protected final DateTime processStart;
    protected final long processStartLong;
    protected final long processDuration;
    protected final int intervalCount;
    protected final Period intervalSize;
    protected final long intervalSizeApproximateLong;


    public AbstractTimeWindowData(DateTime processStart, DateTime processEnd, AggregationIntervalSettings aggregationIntervalSettings) {
        Duration processDuration = new Duration(processStart, processEnd);
        Period intervalSize = aggregationIntervalSettings.getPeriod();

        this.intervalSize = intervalSize;
        this.intervalSizeApproximateLong = intervalSize.toDurationFrom(new Instant(0)).getMillis();
        this.processDuration = processDuration.getMillis();
        this.processStart = processStart;
        this.processStartLong = processStart.getMillis();
        this.intervalCount = bin(processEnd) + 1;
    }

    public abstract Iterable<BinClassCounts> getCounts(Segment segment);

    public abstract Iterable<BinClassCounts> getSubsetCounts(Segment segment, DateTime from, DateTime to);

    public abstract Iterable<BinClassCounts> getSubsetCounts(Segment segment, int from, int to);

    public int bin(DateTime time) {
        long relStart = time.getMillis() - processStartLong;
        return findBin(time, (int) (relStart / intervalSizeApproximateLong), 0); // floor division
    }

    private int findBin(DateTime time, int guess, int count) {
        if (count > 10) {
            System.out.println("Taking over 10 corrections for binning");
            return 0;
        }
        DateTime leftGuess = processStart.plus(intervalSize.multipliedBy(guess));
        DateTime rightGuess = processStart.plus(intervalSize.multipliedBy(guess + 1));
        if (time.isBefore(leftGuess)) return findBin(time, guess - 1, count + 1);
        else if (!time.isBefore(rightGuess)) return findBin(time, guess + 1, count + 1);
        else {
            return guess;
        }
    }

    public abstract ClassCounts.Counts max(Segment segment);

    public int getIntervalCount() {
        return intervalCount;
    }

    public Period getIntervalSize() {
        return intervalSize;
    }

    public abstract boolean containsAll(Collection<Segment> selectedSegments);

    public static class BinClassCounts {
        private int bin;
        private ClassCounts classCounts;

        public BinClassCounts(int bin, ClassCounts classCounts) {
            this.bin = bin;
            this.classCounts = classCounts;
        }

        public ClassCounts getClassCounts() {
            return classCounts;
        }

        public int getBin() {
            return bin;
        }
    }


}
