package org.processmining.performancespectrumintegration.models;

import com.google.common.collect.Maps;

import java.util.Iterator;
import java.util.Map;
import java.util.SortedMap;

public class ClassCounts implements Iterable<Map.Entry<Integer, ClassCounts.Counts>> {

    private final SortedMap<Integer, Counts> internal;

    private ClassCounts(SortedMap<Integer, Counts> internal) {
        this.internal = internal;
    }

    public ClassCounts() {
        internal = Maps.newTreeMap();
    }

    public void incStartingIn(Integer classId) {
        checkKey(classId);
        internal.get(classId).incStartingIn();
    }

    private synchronized void checkKey(int classId) {
        if (!internal.containsKey(classId)) internal.put(classId, new Counts());
    }

    public void incPending(Integer classId) {
        checkKey(classId);
        internal.get(classId).incPending();
    }

    public void incEndingIn(Integer classId) {
        checkKey(classId);
        internal.get(classId).incEndingIn();
    }


    public ClassCounts max(ClassCounts other) {
        SortedMap<Integer, Counts> maxClassCounts = Maps.newTreeMap();
        for (Map.Entry<Integer, Counts> entry : internal.entrySet()) {
            Counts myCounts = entry.getValue();
            Counts otherCounts = other.internal.get(entry.getKey());
            Counts max = myCounts.max(otherCounts);
            maxClassCounts.put(entry.getKey(), max);
        }
        return new ClassCounts(maxClassCounts);
    }

    @Override
    public String toString() {
        return "ClassCounts{" +
                "internal=" + internal +
                '}';
    }

    public Counts sum() {
        return internal.values().stream().reduce(new Counts(), Counts::add);
    }

    public Counts getCountsForClass(Integer classId) {
        return internal.get(classId);
    }

    @Override
    public Iterator<Map.Entry<Integer, Counts>> iterator() {
        return internal.entrySet().iterator();
    }

    public static class Counts {

        private int startingIn, pending, endingIn;

        public Counts() {
            this(0, 0, 0);
        }

        public Counts(int start, int pend, int end) {
            startingIn = start;
            pending = pend;
            endingIn = end;
        }

        public void incStartingIn() {
            ++startingIn;
        }

        public void incPending() {
            ++pending;
        }

        public void incEndingIn() {
            ++endingIn;
        }

        public int getStartingIn() {
            return startingIn;
        }

        public int getPending() {
            return pending;
        }

        public int getEndingIn() {
            return endingIn;
        }

        public Counts max(Counts other) {
            int start = Math.max(startingIn, other.startingIn);
            int pend = Math.max(pending, other.pending);
            int end = Math.max(endingIn, other.endingIn);
            return new Counts(start, pend, end);
        }

        public Counts add(Counts other) {
            int start = startingIn + other.startingIn;
            int pend = pending + other.pending;
            int end = endingIn + other.endingIn;
            return new Counts(start, pend, end);
        }

        public int sum() {
            return startingIn + pending + endingIn;
        }

        public int max() {
            return Math.max(Math.max(startingIn, pending), endingIn);
        }

        @Override
        public String toString() {
            return "Counts{" +
                    "startingIn=" + startingIn +
                    ", pending=" + pending +
                    ", endingIn=" + endingIn +
                    '}';
        }
    }


}
