package org.processmining.performancespectrumintegration.models;

import org.joda.time.DateTime;
import org.processmining.performancespectrumintegration.algorithms.miner.LogAlignmentInfoExtractor;
import org.processmining.performancespectrumintegration.algorithms.miner.SegmentDataMiner;
import org.processmining.performancespectrumintegration.algorithms.segments.Segment;
import org.processmining.performancespectrumintegration.models.settings.AggregationIntervalSettings;
import org.processmining.performancespectrumintegration.models.settings.CaseIdClassification;
import org.processmining.performancespectrumintegration.models.settings.CaseIdFilter;
import org.processmining.performancespectrumintegration.models.settings.DataSettings;
import org.processmining.performancespectrumintegration.util.Tuple2;
import org.processmining.performancespectrumintegration.util.Utils;
import org.processmining.performancespectrumintegration.visualization.controller.ExtraComputationsController;

import java.util.List;
import java.util.Set;
import java.util.function.Supplier;

public class DataController<T extends DataControllable> {

    private final ExtraComputationsController extraComputationsController;
    private final SegmentDataMiner dataMiner;
    private final LogAlignmentInterface logAlignmentInterface;
    private final Supplier<T> dataSourceSource;
    private T dataSource;
    private Set<Segment> lastSelectedSegments;
    private AggregationIntervalSettings lastAggregationIntervalSettings;
    private CaseIdFilter lastFilter;
    private CaseIdClassification lastClassification;
    private Tuple2<DateTime, DateTime> lastTimePair;
    private List<Segment> lastSegmentOrder;
    private TimeWindowData lastTimeWindowData;
    private DataSettings currentSettings;
    private int logSize;

    public DataController(InteractivePSM interactivePSM, Supplier<T> dataSourceSource) {
        this.dataSourceSource = dataSourceSource;
        dataMiner = new SegmentDataMiner(interactivePSM.getLog(), interactivePSM.getAcceptingNet(), interactivePSM.getPnRepResult());
        LogAlignmentInfoExtractor logAlignmentInfoExtractor = new LogAlignmentInfoExtractor(interactivePSM.getLog(), interactivePSM.getPnRepResult());
        logAlignmentInterface = new LogAlignmentInterface(logAlignmentInfoExtractor);
        extraComputationsController = new ExtraComputationsController(logAlignmentInterface);
        dataSource = dataSourceSource.get();
        currentSettings = new DataSettings();
        logSize = interactivePSM.getLog().size();
    }

    public T getDataSource() {
        return dataSource;
    }


    public void prepareCache(Set<Segment> selectedSegments) {
        boolean templates = dataMiner.updateTemplates(selectedSegments);
        boolean cache = dataMiner.updateCachedData(selectedSegments, lastFilter == null ? i -> true : lastFilter);// TODO test if performance and RAM use is feasible
        System.out.println("preemptively cached " + selectedSegments.size() + (templates ? " templates and " : " ") + (cache ? " raw data" : ""));
    }

    public void applyNewSettingsSequentially(DataSettings deltaSettings) throws Exception {
        if (!deltaSettings.isAnySet()) return;

        currentSettings.apply(deltaSettings);

        boolean filterChanged = false;
        final CaseIdFilter filter;
        if (deltaSettings.getFilterSettings() != null) {
            filter = logAlignmentInterface.buildFilter(currentSettings.getFilterSettings()).call();
            System.out.println("Computed new filter");
            filterChanged = true;
        } else {
            filter = lastFilter;
        }

        boolean classificationChanged = false;
        final CaseIdClassification classification;
        if (deltaSettings.getClassificationSettings() != null || filterChanged) {
            classification = logAlignmentInterface.buildClassifier(filter, currentSettings.getClassificationSettings()).call();
            System.out.println("Computed new classifier");
            classificationChanged = true;
        } else
            classification = lastClassification;

        final Set<Segment> selectedSegments;
        if (deltaSettings.getSelectedSegments() != null) {
            selectedSegments = currentSettings.getSelectedSegments();
            boolean updatedTemplates = dataMiner.updateTemplates(selectedSegments);
            System.out.println("did update templates? " + updatedTemplates);
        } else {
            selectedSegments = lastSelectedSegments;
        }

        final List<Segment> segmentOrder;
        if (deltaSettings.getSelectedSegments() != null || deltaSettings.getSegmentOrderSettings() != null) {
            segmentOrder = logAlignmentInterface.buildSegmentOrder(currentSettings.getSegmentOrderSettings(), selectedSegments, i -> true).call();
            System.out.println("computed new segment order");
        } else
            segmentOrder = lastSegmentOrder;

        boolean updateCachedData = dataMiner.updateCachedData(selectedSegments, filter);
        System.out.println("did update cached data? " + updateCachedData);

        boolean aggregationIntervalSettingsChanged = false;
        final AggregationIntervalSettings aggregationIntervalSettings;
        if (deltaSettings.getAggregationIntervalSettings() != null || updateCachedData || filterChanged || classificationChanged || lastTimeWindowData == null || !lastTimeWindowData.containsAll(selectedSegments)) {
            aggregationIntervalSettings = currentSettings.getAggregationIntervalSettings();
            aggregationIntervalSettingsChanged = true;
        } else aggregationIntervalSettings = lastAggregationIntervalSettings;

        Tuple2<DateTime, DateTime> timePair;
        if (filterChanged || aggregationIntervalSettingsChanged)
            timePair = logAlignmentInterface.buildProcessDuration(filter).call();
        else
            timePair = lastTimePair;

        if (aggregationIntervalSettingsChanged && currentSettings.getAggregationIntervalSettings().isNormalizeProcessStart() && timePair._1 != null) {
            timePair._1 = Utils.normalizeToPeriod(timePair._1, currentSettings.getAggregationIntervalSettings());
        }

        final TimeWindowData timeWindowData;
        if (aggregationIntervalSettingsChanged) {
            if (timePair._1 == null) timeWindowData = null;
            else
                timeWindowData = new TimeWindowData(dataMiner.getSegmentData(), selectedSegments, classification, timePair._1, timePair._2, aggregationIntervalSettings);
            System.out.println("computed new timewindowdata");
        } else {
            timeWindowData = lastTimeWindowData;
        }

        lastFilter = filter;
        lastClassification = classification;
        lastTimePair = timePair;
        lastAggregationIntervalSettings = aggregationIntervalSettings;
        lastSelectedSegments = selectedSegments;
        lastSegmentOrder = segmentOrder;
        lastTimeWindowData = timeWindowData;

        T newDataSource = dataSourceSource.get();
        newDataSource.setProcessTimes(timePair._1, timePair._2);
        newDataSource.setCaseNumbers(logSize, classification.getDefinitionSpaceSize());
        newDataSource.setClassification(classification);
        newDataSource.setSegmentData(dataMiner.getSegmentData());
        newDataSource.setSegments(selectedSegments);
        newDataSource.setSegmentOrder(segmentOrder);
        newDataSource.setTimeWindowData(timeWindowData);
        dataSource = newDataSource;
    }

    public ExtraComputationsController getExtraComputationsController() {
        return extraComputationsController;
    }

}
