package org.processmining.performancespectrumintegration.models;

import com.google.common.collect.Sets;
import org.jfree.data.category.CategoryDataset;
import org.jfree.data.statistics.HistogramDataset;
import org.joda.time.DateTime;
import org.processmining.performancespectrumintegration.algorithms.miner.LogAlignmentInfoExtractor;
import org.processmining.performancespectrumintegration.algorithms.segments.Segment;
import org.processmining.performancespectrumintegration.models.settings.*;
import org.processmining.performancespectrumintegration.util.Tuple2;

import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.concurrent.Callable;

public class LogAlignmentInterface {

    private final LogAlignmentInfoExtractor logAlignmentInfoExtractor;

    public LogAlignmentInterface(LogAlignmentInfoExtractor logAlignmentInfoExtractor) {
        this.logAlignmentInfoExtractor = logAlignmentInfoExtractor;
    }


    public Callable<CaseIdFilter> buildFilter(FilterSettings filterSettings) {
        Callable<CaseIdFilter> result = null;
        switch (filterSettings.getType()) {
            case VariantPercentile:
                PercentageFilterSettings ips1 = (PercentageFilterSettings) filterSettings;
                result = () -> logAlignmentInfoExtractor.getVariantFilter(ips1.getPercentage(), ips1.isInverted());
                break;
            case CaseDurationThreshold:
                DurationThresholdFilterSettings ips2 = (DurationThresholdFilterSettings) filterSettings;
                result = () -> logAlignmentInfoExtractor.getDurationThresholdFilter(ips2.getThresholdDuration(), ips2.isInverted());
                break;
            case CaseDurationPercentile:
                PercentageFilterSettings ips3 = (PercentageFilterSettings) filterSettings;
                result = () -> logAlignmentInfoExtractor.getDurationPercentileFilter(ips3.getPercentage(), ips3.isInverted());
                break;
            case CaseIds:
                CaseIdFilterSettings ips4 = (CaseIdFilterSettings) filterSettings;
                result = () -> new SetBasedFilter(((CaseIdFilterSettings) filterSettings).getCaseIds(), ips4.isInverted());
                break;
            case None:
            default:
                result = logAlignmentInfoExtractor::buildNoFilter;
                break;
        }
        return result;
    }

    public Callable<CaseIdClassification> buildClassifier(CaseIdFilter selectedCaseIds, ClassificationSettings classificationSettings) {
        Callable<CaseIdClassification> result = null;
        switch (classificationSettings.getType()) {
            case VariantPercentile:
                PercentileClassificationSettings percentileClassificationSettings = (PercentileClassificationSettings) classificationSettings;
                result = () -> logAlignmentInfoExtractor.getBinaryVariantClassifier(selectedCaseIds, percentileClassificationSettings.getTopPercentage());
                break;
            case CaseDurationPercentile:
                PercentileClassificationSettings percentileClassificationSettings2 = (PercentileClassificationSettings) classificationSettings;
                result = () -> logAlignmentInfoExtractor.getBinaryDurationClassifier(selectedCaseIds, percentileClassificationSettings2.getTopPercentage());
                break;
            case CaseDurationThreshold:
                DurationThresholdClassificationSettings durationThresholdClassificationSettings = (DurationThresholdClassificationSettings) classificationSettings;
                result = () -> logAlignmentInfoExtractor.getDurationThresholdClassifier(selectedCaseIds, durationThresholdClassificationSettings.getThresholdDuration());
                break;
            case CaseDurationClasses:
                DurationClassesClassificationSettings durationClassesClassificationSettings = (DurationClassesClassificationSettings) classificationSettings;
                result = () -> logAlignmentInfoExtractor.getDurationClassesClassifier(selectedCaseIds, durationClassesClassificationSettings.getClasses());
                break;
            case None:
            default:
                result = () -> logAlignmentInfoExtractor.buildNoClassifier(selectedCaseIds);
                break;
        }
        return result;
    }

    public Callable<List<Segment>> buildSegmentOrder(SegmentOrderSettings segmentOrderSettings, Set<Segment> selectedSegments, CaseIdFilter selectedCaseIds) {
        if (segmentOrderSettings.usePredefined()) return () -> {
            List<Segment> predefinedSegmentOrder = segmentOrderSettings.getPredefinedOrder();
            Set<Segment> remaining = Sets.newHashSet(selectedSegments);
            List<Segment> result = new LinkedList<>();
            for (Segment segment : predefinedSegmentOrder) {
                if (selectedSegments.contains(segment)) {
                    result.add(segment);
                    remaining.remove(segment);
                }
            }
            result.addAll(remaining);
            return result;
        };
        else
            return () -> logAlignmentInfoExtractor.getSegmentOrder(selectedSegments, selectedCaseIds);
    }

    public Callable<Tuple2<DateTime, DateTime>> buildProcessDuration(CaseIdFilter selectedCaseIds) {
        return () -> logAlignmentInfoExtractor.getProcessDuration(selectedCaseIds);
    }

    public Tuple2<CategoryDataset, List<String>> getVariantDistribution(CaseIdFilter caseIdFilter) {
        return logAlignmentInfoExtractor.getVariantDistribution(caseIdFilter);
    }

    public HistogramDataset getDurationHistogramDataset(CaseIdFilter caseIdFilter) {
        return logAlignmentInfoExtractor.getDurationHistogramDataset(caseIdFilter);
    }
}
