package org.processmining.performancespectrumintegration.models;

import com.google.common.collect.Maps;
import org.joda.time.DateTime;
import org.processmining.performancespectrumintegration.algorithms.miner.SegmentData;
import org.processmining.performancespectrumintegration.algorithms.segments.Segment;
import org.processmining.performancespectrumintegration.algorithms.segments.SegmentOccurrence;
import org.processmining.performancespectrumintegration.models.settings.AggregationIntervalSettings;
import org.processmining.performancespectrumintegration.models.settings.CaseIdClassification;

import java.util.*;

public class TimeWindowData extends AbstractTimeWindowData {

    private final Map<Segment, SortedMap<Integer, ClassCounts>> timeWindowCounts;
    private final Map<Segment, ClassCounts.Counts> maxMap;

    public TimeWindowData(SegmentData segmentData, Set<Segment> selectedSegments, CaseIdClassification classification, DateTime processStart, DateTime processEnd, AggregationIntervalSettings aggregationIntervalSettings) {
        super(processStart, processEnd, aggregationIntervalSettings);
        timeWindowCounts = Maps.newHashMap();

        for (Segment segment : selectedSegments) {
            for (SegmentOccurrence so : segmentData.get(segment)) {
                Integer caseId = so.getCaseId();
                Integer classId = classification.getClass(caseId);

                int s = bin(so.getStartTime());
                int e = bin(so.getEndTime());

                check(segment, s).incStartingIn(classId);
                for (int i = s; i <= e; i++) {
                    check(segment, i).incPending(classId);
                }
                check(segment, e).incEndingIn(classId);

            }
        }

        maxMap = Maps.transformValues(timeWindowCounts, map -> map.values().stream().map(ClassCounts::sum).reduce(ClassCounts.Counts::max).orElseGet(ClassCounts.Counts::new));

    }

    @Override
    public Iterable<BinClassCounts> getCounts(Segment segment) {
        return new WrapperIterable(timeWindowCounts.get(segment));
    }

    @Override
    public Iterable<BinClassCounts> getSubsetCounts(Segment segment, DateTime from, DateTime to) {
        return getSubsetCounts(segment, bin(from), bin(to));
    }

    @Override
    public Iterable<BinClassCounts> getSubsetCounts(Segment segment, int from, int to) {
        if (!timeWindowCounts.containsKey(segment)) return null;
        return new WrapperIterable(timeWindowCounts.get(segment).subMap(from, to + 1)); // because to is exclusive
    }

    @Override
    public ClassCounts.Counts max(Segment segment) {
        if (!timeWindowCounts.containsKey(segment)) return new ClassCounts.Counts();
        return maxMap.get(segment);
    }

    private ClassCounts check(Segment segment, int index) {
        if (!timeWindowCounts.containsKey(segment)) {
            timeWindowCounts.put(segment, Maps.newTreeMap());
        }
        SortedMap<Integer, ClassCounts> map = timeWindowCounts.get(segment);
        if (!map.containsKey(index)) {
            map.put(index, new ClassCounts());
        }
        return map.get(index);
    }

    @Override
    public boolean containsAll(Collection<Segment> selectedSegments) {
        return timeWindowCounts.keySet().containsAll(selectedSegments);
    }

    private class WrapperIterable implements Iterable<BinClassCounts> {

        private final SortedMap<Integer, ClassCounts> map;

        public WrapperIterable(SortedMap<Integer, ClassCounts> map) {
            this.map = map;
        }

        @Override
        public Iterator<BinClassCounts> iterator() {
            return new Iterator<BinClassCounts>() {
                private Iterator<Map.Entry<Integer, ClassCounts>> internal = map.entrySet().iterator();

                @Override
                public boolean hasNext() {
                    return internal.hasNext();
                }

                @Override
                public BinClassCounts next() {
                    Map.Entry<Integer, ClassCounts> next = internal.next();
                    return new BinClassCounts(next.getKey(), next.getValue());
                }
            };
        }
    }

}
