package org.processmining.performancespectrumintegration.mypsm;

import com.fluxicon.slickerbox.factory.SlickerFactory;
import org.joda.time.Period;
import org.processmining.performancespectrumintegration.models.settings.DataSettings;
import org.processmining.performancespectrumintegration.mypsm.update.PSMUpdateListener;
import org.processmining.performancespectrumintegration.mypsm.update.PSMUpdateManager;
import org.processmining.performancespectrumintegration.mypsm.update.PSMVisualSettings;
import org.processmining.performancespectrumintegration.mypsm.update.PSMVisualSettings.InbetweenTick;
import org.processmining.performancespectrumintegration.util.Tuple2;

import javax.swing.*;
import java.awt.*;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Hashtable;
import java.util.stream.IntStream;

public class PSMControlBar extends JPanel implements PSMUpdateListener {

    private static final int VIEW_VALUES = 300;
    private final PSMUpdateManager psmUpdateManager;

    private final JSlider viewSlider;
    private final JSlider xScaleSlider;
    private final JCheckBox helplinesCheckbox;
    private final JComboBox<InbetweenTick> inbetweenTickComboBox;
    private final JCheckBox backgroundPatternCheckbox;
    private final ViewScalePanel viewScalePanel;
    private final JCheckBox approxLinesCheckbox;
    private final JSlider yScaleSlider;
    private final JComboBox<PSMVisualSettings.LineDensity> lineDensityComboBox;
    private boolean isInternal;

    public PSMControlBar(PSMUpdateManager psmUpdateManager, SegmentsListPanel segmentsListPanel) {
        this.psmUpdateManager = psmUpdateManager;
        this.isInternal = false;

        PSMVisualSettings visualSettings = psmUpdateManager.getVisualSettings();

        setLayout(new GridBagLayout());

        GridBagConstraints c = new GridBagConstraints();
        c.weightx = 0.1;
        c.weighty = 0;
        c.ipadx = 5;
        c.ipady = 5;
        c.anchor = GridBagConstraints.WEST;
        c.fill = GridBagConstraints.NONE;

        JButton reorderButton = SlickerFactory.instance().createButton("reset");
        reorderButton.addActionListener(e -> {
            visualSettings.reset();
            segmentsListPanel.reset();
            psmUpdateManager.visualSettingsChanged();
        });

        c.gridheight = 4;
        c.gridx = 0;
        c.gridy = 0;
        add(reorderButton, c);
        c.weightx = 1;
        c.weighty = 1;

        xScaleSlider = SlickerFactory.instance().createSlider(JSlider.HORIZONTAL);
        Tuple2<Double, Double> range = visualSettings.getExponentialScale().getExponentRange();
        xScaleSlider.setMinimum(range._1.intValue());
        xScaleSlider.setMaximum(range._2.intValue());
        xScaleSlider.setValue((int) visualSettings.getExponentialScale().getExponent());
        xScaleSlider.addChangeListener(e -> {
            if (isInternal) return;

            int value = xScaleSlider.getValue();
            visualSettings.getExponentialScale().setExponent(value);
            psmUpdateManager.visualSettingsChanged();
        });

        c.gridheight = 1;
        c.gridy = 0;
        c.gridx = 1;
        c.weightx = 0.1;
        JLabel zoomLabel = SlickerFactory.instance().createLabel("Zoom");
        add(zoomLabel, c);
        c.weightx = 1;
        c.fill = GridBagConstraints.HORIZONTAL;
        c.gridwidth = 3;
        c.gridx = 2;
        add(xScaleSlider, c);
        c.gridwidth = 1;
        c.fill = GridBagConstraints.NONE;


        yScaleSlider = SlickerFactory.instance().createSlider(JSlider.HORIZONTAL);
        yScaleSlider.setMinimum((int) (PSMVisualSettings.MIN_YSCALE * 100));
        yScaleSlider.setMaximum(PSMVisualSettings.MAX_YSCALE * 100);
        yScaleSlider.setValue((int) (visualSettings.getYScale() * 100));
        yScaleSlider.addChangeListener(e -> {
            if (isInternal) return;

            int value = yScaleSlider.getValue();
            visualSettings.setYScale((double) value / 100.0);
            psmUpdateManager.visualSettingsChanged();
        });

        c.gridy = 0;
        c.gridx = 5;
        c.weightx = 0.1;
        add(SlickerFactory.instance().createLabel("Segment Height"), c);
        c.weightx = 1;
        c.gridx = 6;
        c.fill = GridBagConstraints.HORIZONTAL;
        add(yScaleSlider, c);
        c.fill = GridBagConstraints.NONE;

        viewSlider = SlickerFactory.instance().createSlider(JSlider.HORIZONTAL);
        viewSlider.setMinimum(0);
        viewSlider.setMaximum(VIEW_VALUES);
        viewSlider.setValue((int) (visualSettings.getViewAnchor() * VIEW_VALUES));
        viewSlider.addChangeListener(e -> {
            if (isInternal) return;

            int value = viewSlider.getValue();
            visualSettings.setViewAnchor((double) value / 300.0);
            psmUpdateManager.visualSettingsChanged();
        });

        c.gridy = 1;
        c.gridx = 1;
        c.weightx = 0.1;
        JLabel viewLabel = SlickerFactory.instance().createLabel("View");
        viewLabel.setToolTipText("can also be changed by dragging or scrolling while holding shift on the segment itself");
        add(viewLabel, c);
        c.weightx = 1;
        c.gridwidth = 3;
        c.gridx = 2;
        c.fill = GridBagConstraints.HORIZONTAL;
        add(viewSlider, c);
        c.gridwidth = 1;
        c.fill = GridBagConstraints.NONE;


        viewScalePanel = new ViewScalePanel(psmUpdateManager);

        c.gridy = 1;
        c.gridx = 5;
        c.weightx = 0.1;
        JLabel viewScaleLabel = SlickerFactory.instance().createLabel("View Position");
        add(viewScaleLabel, c);
        c.weightx = 1;
        c.gridx = 6;
        c.fill = GridBagConstraints.HORIZONTAL;
        add(viewScalePanel, c);
        c.fill = GridBagConstraints.NONE;


        lineDensityComboBox = SlickerFactory.instance().createComboBox(PSMVisualSettings.LineDensity.values());
        String text = "determines how many lines of a segment will actually be drawn (increases performance)";
        lineDensityComboBox.setToolTipText(text);
        lineDensityComboBox.addActionListener(e -> {
            if (isInternal) return;

            PSMVisualSettings.LineDensity selectedItem = (PSMVisualSettings.LineDensity) lineDensityComboBox.getSelectedItem();
            if (selectedItem != null) {
                visualSettings.setLineDensity(selectedItem);
                psmUpdateManager.visualSettingsChanged();
            }
        });

        c.ipady = 0;
        c.gridy = 2;
        c.gridx = 1;
        c.gridheight = 2;
        c.weightx = 0.1;
        JLabel lineDensityLabel = SlickerFactory.instance().createLabel("Line Density");
        lineDensityLabel.setToolTipText(text);
        add(lineDensityLabel, c);
        c.weightx = 1;
        c.gridx = 2;
        add(lineDensityComboBox, c);


        inbetweenTickComboBox = SlickerFactory.instance().createComboBox(InbetweenTick.values());
        inbetweenTickComboBox.setSelectedItem(visualSettings.getInbetweenTick());
        inbetweenTickComboBox.addActionListener(e -> {
            if (isInternal) return;

            InbetweenTick selected = (InbetweenTick) inbetweenTickComboBox.getSelectedItem();
            if (selected != null) {
                visualSettings.setInbetweenTick(selected);
                psmUpdateManager.visualSettingsChanged();
            }
        });

        c.gridy = 2;
        c.gridx = 3;
        c.weightx = 0.1;
        JLabel patternIntervalLabel = SlickerFactory.instance().createLabel("Pattern Interval");
        add(patternIntervalLabel, c);
        c.weightx = 1;
        c.gridx = 4;
        add(inbetweenTickComboBox, c);


        backgroundPatternCheckbox = SlickerFactory.instance().createCheckBox("show pattern", visualSettings.isShowBackgroundPattern());
        backgroundPatternCheckbox.addActionListener(e -> {
            if (isInternal) return;

            visualSettings.setShowBackgroundPattern(backgroundPatternCheckbox.isSelected());
            psmUpdateManager.visualSettingsChanged();
        });

        c.gridheight = 1;
        c.gridwidth = 1;
        c.weightx = 0.1;
        c.gridy = 2;
        c.gridx = 5;
        add(backgroundPatternCheckbox, c);


        helplinesCheckbox = SlickerFactory.instance().createCheckBox("show dotted lines", visualSettings.isShowHelpLines());
        helplinesCheckbox.addActionListener(e -> {
            if (isInternal) return;

            visualSettings.setShowHelpLines(helplinesCheckbox.isSelected());
            psmUpdateManager.visualSettingsChanged();
        });

        c.gridy = 2;
        c.gridx = 6;
        add(helplinesCheckbox, c);


        approxLinesCheckbox = SlickerFactory.instance().createCheckBox("show only starting lines", visualSettings.isSubsetApproximation());
        approxLinesCheckbox.setToolTipText("only lines starting in the current view are drawn (increases performance)");
        approxLinesCheckbox.addActionListener(e -> {
            if (isInternal) return;

            visualSettings.setSubsetApproximation(approxLinesCheckbox.isSelected());
            psmUpdateManager.visualSettingsChanged();
        });

        c.gridy = 3;
        c.gridx = 5;
        c.gridwidth = 2;
        add(approxLinesCheckbox, c);

        /*
        for (Component component : getComponents()) {
            if (component instanceof JComponent)
                ((JComponent) component).setBorder(new LineBorder(Color.red));
        }
         */

        psmUpdateManager.addListener(this);
    }

    private JPanel packInPanel(JComponent... jcs) {
        JPanel jp = new JPanel();
        for (JComponent jc : jcs) {
            jp.add(jc);
        }
        return jp;
    }

    @Override
    public void onVisualSettingsChanged() {
        SwingUtilities.invokeLater(() -> {
            isInternal = true;
            PSMVisualSettings vs = psmUpdateManager.getVisualSettings();

            viewSlider.setValue((int) (VIEW_VALUES * vs.getViewAnchor()));
            xScaleSlider.setValue((int) vs.getExponentialScale().getExponent());
            yScaleSlider.setValue((int) (100 * vs.getYScale()));

            lineDensityComboBox.setSelectedItem(vs.getLineDensity());

            approxLinesCheckbox.setSelected(vs.isSubsetApproximation());

            helplinesCheckbox.setSelected(vs.isShowHelpLines());

            inbetweenTickComboBox.setSelectedItem(vs.getInbetweenTick());

            backgroundPatternCheckbox.setSelected(vs.isShowBackgroundPattern());

            viewSlider.repaint();
            xScaleSlider.repaint();
            yScaleSlider.repaint();
            viewScalePanel.repaint();
            lineDensityComboBox.repaint();
            inbetweenTickComboBox.repaint();
            approxLinesCheckbox.repaint();
            backgroundPatternCheckbox.repaint();
            helplinesCheckbox.repaint();

            SwingUtilities.invokeLater(() -> {
                isInternal = false;
            });
        });
    }

    @Override
    public void onDataSourceChanged(DataSettings deltaSettings) {
        SwingUtilities.invokeLater(() -> {
            isInternal = true;

            viewScalePanel.repaint();

            if (deltaSettings.getAggregationIntervalSettings() != null) {
                Period intervalSize = psmUpdateManager.getDataSource().getTimeWindowData().getIntervalSize();
                InbetweenTick[] inbetweenTicks = InbetweenTick.valuesShorterThan(intervalSize);
                ComboBoxModel<InbetweenTick> model = new DefaultComboBoxModel<>(inbetweenTicks);
                InbetweenTick oldSelection = (InbetweenTick) inbetweenTickComboBox.getSelectedItem();
                inbetweenTickComboBox.setModel(model);
                if (oldSelection != null && Arrays.stream(inbetweenTicks).anyMatch(it -> it == oldSelection))
                    model.setSelectedItem(oldSelection);

                InbetweenTick selectedItem = (InbetweenTick) inbetweenTickComboBox.getSelectedItem();
                psmUpdateManager.getVisualSettings().setInbetweenTick(selectedItem);
                psmUpdateManager.visualSettingsChanged();
            }

            SwingUtilities.invokeLater(() -> {
                isInternal = false;
            });

        });

    }

    private static class ResolutionSlider extends JSlider {
        private static final long serialVersionUID = -1195270044097152629L;
        private static final Integer[] VALUES = {20, 10, 5, 3, 2, 1};
        private static final Hashtable<Integer, JLabel> LABELS = new Hashtable<>();

        static {
            for (int i = 0; i < VALUES.length; ++i) {
                LABELS.put(i, new JLabel(VALUES[i].toString()));
            }
        }

        public ResolutionSlider() {
            super(0, VALUES.length - 1, VALUES.length - 1);
            setLabelTable(LABELS);
            setPaintTicks(true);
            setPaintLabels(true);
            setSnapToTicks(true);
            setMajorTickSpacing(1);
        }

        public int getDomainValue() {
            return VALUES[getValue()];
        }

        public void setDomainValue(int i) {
            Integer n = IntStream.range(0, VALUES.length).boxed().sorted(Comparator.comparing(v -> (int) Math.abs(VALUES[v] - i))).toArray(Integer[]::new)[0];
            setValue(n);
        }
    }

}
