package org.processmining.performancespectrumintegration.mypsm;

import com.google.common.collect.BiMap;
import com.google.common.collect.HashBiMap;
import com.google.common.collect.Maps;
import org.processmining.performancespectrumintegration.algorithms.segments.Segment;
import org.processmining.performancespectrumintegration.models.MyDataSource;
import org.processmining.performancespectrumintegration.models.settings.DataSettings;
import org.processmining.performancespectrumintegration.mypsm.segmentpanel.SegmentPanel;
import org.processmining.performancespectrumintegration.mypsm.update.PSMUpdateListener;
import org.processmining.performancespectrumintegration.mypsm.update.PSMUpdateManager;
import org.processmining.performancespectrumintegration.mypsm.update.SegmentOrderConnection;

import javax.swing.*;
import java.awt.*;
import java.util.ArrayList;
import java.util.Map;
import java.util.SortedMap;

public class SegmentsListPanel extends JScrollPane implements PSMUpdateListener {

    private final Map<Segment, SegmentPanel> segmentPanelMap;
    private final PSMUpdateManager psmUpdateManager;
    private final JPanel content;
    private final BiMap<SegmentPanel, Integer> panelOrder;

    public SegmentsListPanel(PSMUpdateManager psmUpdateManager) {
        super(ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS, ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
        this.psmUpdateManager = psmUpdateManager;
        panelOrder = HashBiMap.create();
        segmentPanelMap = Maps.newHashMap();
        content = new JPanel();

        content.setLayout(new GridBagLayout());

        setViewportView(content);
        addSegmentPanels();

        psmUpdateManager.addListener(this);
    }

    private void addSegmentPanels() {
        SwingUtilities.invokeLater(() -> {
            MyDataSource dataSource = psmUpdateManager.getDataSource();

            if (dataSource.getSelectedSegments() == null) return;

            GridBagConstraints c = new GridBagConstraints();
            c.fill = GridBagConstraints.HORIZONTAL;
            c.gridy = 0;
            c.weightx = 1;
            c.anchor = GridBagConstraints.NORTH;

            segmentPanelMap.keySet().retainAll(dataSource.getSegmentOrder());
            for (Segment segment : dataSource.getSegmentOrder()) {
                SegmentPanel panel;
                if (segmentPanelMap.containsKey(segment)) {
                    panel = segmentPanelMap.get(segment);
                    panel.updateDataSettings(DataSettings.defaultSettings());
                } else {
                    panel = new SegmentPanel(segment, psmUpdateManager, this);
                    segmentPanelMap.put(segment, panel);
                }
                panelOrder.put(panel, c.gridy);
                content.add(panel, c);

                c.gridy++;
            }

            c.fill = GridBagConstraints.VERTICAL;
            c.weighty = 1;
            content.add(new JPanel(), c);

            updateSegmentOrderConnection();

            revalidate();
            repaint();
        });
    }

    private void updateSegmentPanels(DataSettings deltaSettings) {
        SwingUtilities.invokeLater(() -> {
            for (SegmentPanel panel : segmentPanelMap.values()) {
                panel.updateDataSettings(deltaSettings);
            }
        });
    }

    private void updateSegmentPanels() {
        SwingUtilities.invokeLater(() -> {
            for (SegmentPanel panel : segmentPanelMap.values()) {
                panel.updateVisualSettings();
                panel.invalidate();
            }
            revalidate();
            repaint();
        });
    }

    private void updateSegmentOrderConnection() {
        SegmentOrderConnection segmentOrderConnection = psmUpdateManager.getSegmentOrderConnection();
        SortedMap<Integer, Segment> temp = Maps.newTreeMap();
        for (Map.Entry<SegmentPanel, Integer> entry : panelOrder.entrySet()) {
            temp.put(entry.getValue(), entry.getKey().getSegment());
        }
        ArrayList<Segment> order = new ArrayList<>(temp.values());
        segmentOrderConnection.setSegmentOrder(order);
    }

    private void updateSegmentHighlighting() {
        SwingUtilities.invokeLater(() -> {
            for (SegmentPanel panel : segmentPanelMap.values()) {
                panel.updateHighlighting();
            }
        });
    }

    @Override
    public void onPlaceSegmentConnectionChanged() {
        updateSegmentHighlighting();
    }

    @Override
    public void onMarkedCaseConnectionChanged() {
        updateSegmentPanels();
    }

    @Override
    public void onVisualSettingsChanged() {
        updateSegmentPanels();
    }

    @Override
    public void onDataSourceChanged(DataSettings deltaSettings) {
        if (deltaSettings.getSelectedSegments() != null || deltaSettings.getSegmentOrderSettings() != null) {
            remake();
        } else {
            updateSegmentPanels(deltaSettings);
        }
    }

    private void remake() {
        content.removeAll();
        //segmentPanelMap.clear();
        panelOrder.clear();
        addSegmentPanels();
    }

    @Override
    public void onColorMapChanged() {
        updateSegmentPanels();
    }

    public void reset() {
        segmentPanelMap.clear();
        remake();
    }

    private void swap(SegmentPanel one, SegmentPanel two) {
        SwingUtilities.invokeLater(() -> {
            int oneIndex = panelOrder.get(one);
            int twoIndex = panelOrder.get(two);

            panelOrder.remove(one);
            panelOrder.remove(two);

            panelOrder.put(one, twoIndex);
            panelOrder.put(two, oneIndex);

            //content.remove(oneIndex);
            //content.remove(twoIndex);

            GridBagConstraints c = new GridBagConstraints();
            c.fill = GridBagConstraints.HORIZONTAL;
            c.anchor = GridBagConstraints.NORTH;
            c.weightx = 1;
            c.gridy = oneIndex;
            content.add(two, c, oneIndex);
            c.gridy = twoIndex;
            content.add(one, c, twoIndex);

            updateSegmentOrderConnection();

            revalidate();
            repaint();
        });
    }

    public void moveUp(SegmentPanel segmentPanel) {
        Integer id = panelOrder.get(segmentPanel);
        if (id == 0) return;
        SegmentPanel above = panelOrder.inverse().get(id - 1);
        swap(segmentPanel, above);
    }

    public void moveDown(SegmentPanel segmentPanel) {
        Integer id = panelOrder.get(segmentPanel);
        if (id >= panelOrder.size() - 1) return;
        SegmentPanel below = panelOrder.inverse().get(id + 1);
        swap(segmentPanel, below);
    }

}
