package org.processmining.performancespectrumintegration.mypsm.segmentpanel;

import org.joda.time.Period;
import org.processmining.performancespectrumintegration.algorithms.segments.Segment;
import org.processmining.performancespectrumintegration.models.AbstractTimeWindowData;
import org.processmining.performancespectrumintegration.models.ClassCounts;
import org.processmining.performancespectrumintegration.models.MyDataSource;
import org.processmining.performancespectrumintegration.models.TimeWindowData;
import org.processmining.performancespectrumintegration.mypsm.PSMColorMap;
import org.processmining.performancespectrumintegration.mypsm.ViewWindowTools;
import org.processmining.performancespectrumintegration.mypsm.update.PSMAggregationSettings;
import org.processmining.performancespectrumintegration.mypsm.update.PSMUpdateManager;
import org.processmining.performancespectrumintegration.mypsm.update.PSMVisualSettings;

import java.awt.*;
import java.util.Map;
import java.util.function.Function;

public class SegmentBarsPanel extends SegmentDataPanel {

    private static final int MIN__DASHED_DIST = 40;
    private final Segment segment;

    public SegmentBarsPanel(Segment segment, PSMUpdateManager psmUpdateManager) {
        super(psmUpdateManager);
        this.segment = segment;
    }

    @Override
    protected void paintComponent(Graphics g) {
        super.paintComponent(g);

        int width = getWidth();
        int height = getHeight();

        Graphics2D graphics2D = (Graphics2D) g;
        graphics2D.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        graphics2D.clipRect(0, 0, width, height);

        PSMVisualSettings visualSettings = psmUpdateManager.getVisualSettings();
        PSMColorMap psmColorMap = psmUpdateManager.getPsmColorMap();
        MyDataSource dataSource = psmUpdateManager.getDataSource();

        if (dataSource.isComplete()) {
            ViewWindowTools.WindowParams viewWindow = ViewWindowTools.getWindow(visualSettings, dataSource.getProcessStart(), dataSource.getProcessDuration(), width);

            TimeWindowData timeWindowData = dataSource.getTimeWindowData();
            Period intervalSize = timeWindowData.getIntervalSize();

            Iterable<AbstractTimeWindowData.BinClassCounts> subsetCounts = timeWindowData.getSubsetCounts(segment, viewWindow.getWindowStart(), viewWindow.getWindowEnd());

            if (subsetCounts != null) {
                PSMAggregationSettings psmAggregationSettings = psmUpdateManager.getPsmAggregationSettings();
                PSMAggregationSettings.AggregationVersion version = psmAggregationSettings.getAggregationVersion();

                Function<ClassCounts.Counts, Integer> getter = ClassCounts.Counts::getStartingIn;
                switch (version) {
                    case StartingIn:
                        getter = ClassCounts.Counts::getStartingIn;
                        break;
                    case ActiveIn:
                        getter = ClassCounts.Counts::getPending;
                        break;
                    case EndingIn:
                        getter = ClassCounts.Counts::getEndingIn;
                        break;
                }

                int max = getter.apply(timeWindowData.max(segment));

                double yScale = (double) height / max;

                for (AbstractTimeWindowData.BinClassCounts binClassCounts : subsetCounts) {
                    int index = binClassCounts.getBin();

                    int left = getInterval(viewWindow, intervalSize, index)._2;
                    int right = getInterval(viewWindow, intervalSize, index + 1)._2;

                    int padding = (int) ((right - left) * 0.1d);
                    int length = right - left - padding;
                    int x = left + padding / 2;

                    int acc = 0;

                    for (Map.Entry<Integer, ClassCounts.Counts> countsEntry : binClassCounts.getClassCounts()) {
                        Integer classId = countsEntry.getKey();
                        if (!psmColorMap.isActive(classId)) continue;

                        graphics2D.setColor(psmColorMap.getColor(classId));

                        ClassCounts.Counts counts = countsEntry.getValue();

                        int count = (int) (yScale * getter.apply(counts));

                        graphics2D.fillRect(x, height - count - acc, length, count);

                        acc += count;
                    }
                }
                graphics2D.setColor(Color.black);

                if (visualSettings.isShowHelpLines())
                    drawHelpLines(graphics2D, dataSource.getTimeWindowData(), viewWindow);

                int labelDistance = Math.max(10, (height - 20) / 10);
                int lastLabel = -1;
                for (int i = 0; i <= height - 30; i += labelDistance) {
                    int label = (int) (i / yScale);
                    if (label != lastLabel) {
                        int y = height - 1 - i;
                        graphics2D.drawLine(0, y, 5, y);
                        graphics2D.drawString(String.valueOf(label), 5, y);
                        lastLabel = label;
                    }
                }
            } else
                drawX(graphics2D);

        } else
            drawX(graphics2D);


        drawFrame(graphics2D, segment.toString() + " " + dataSource.getSegmentReliability(segment).toShortString(), width, height);

    }


}
