package org.processmining.performancespectrumintegration.mypsm.segmentpanel;

import org.joda.time.DateTime;
import org.joda.time.Instant;
import org.joda.time.Period;
import org.joda.time.format.DateTimeFormat;
import org.joda.time.format.DateTimeFormatter;
import org.processmining.performancespectrumintegration.models.MyDataSource;
import org.processmining.performancespectrumintegration.models.TimeWindowData;
import org.processmining.performancespectrumintegration.models.settings.DataSettings;
import org.processmining.performancespectrumintegration.mypsm.ViewWindowTools;
import org.processmining.performancespectrumintegration.mypsm.ViewWindowTools.WindowParams;
import org.processmining.performancespectrumintegration.mypsm.update.PSMUpdateManager;
import org.processmining.performancespectrumintegration.mypsm.update.PSMVisualSettings;
import org.processmining.performancespectrumintegration.util.ToStrings;
import org.processmining.performancespectrumintegration.util.TriConsumer;
import org.processmining.performancespectrumintegration.util.Tuple2;

import javax.imageio.ImageIO;
import javax.swing.*;
import javax.swing.filechooser.FileNameExtensionFilter;
import java.awt.*;
import java.awt.event.*;
import java.awt.geom.AffineTransform;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.concurrent.ExecutionException;
import java.util.function.BiConsumer;

import static org.processmining.performancespectrumintegration.mypsm.BetterPSMTimeScale.BIG_TICK_ANNOTATION_PADDING;

public class SegmentDataPanel extends UpdatableJPanel implements MouseMotionListener, MouseListener, MouseWheelListener {

    protected PSMUpdateManager psmUpdateManager;
    protected Color accent = new Color(0xE0E0E0);
    private Integer dragStart;
    private Integer dragPosition;
    private Boolean draggedPositive;

    public SegmentDataPanel(PSMUpdateManager psmUpdateManager) {
        this.psmUpdateManager = psmUpdateManager;

        addMouseWheelListener(this);
        addMouseMotionListener(this);
        addMouseListener(this);
    }

    public static Tuple2<DateTime, Integer> getInterval(WindowParams viewWindow, Period intervalSize, int i) {
        DateTime dt = viewWindow.getProcessStart();
        DateTime tickTime = dt.plus(intervalSize.multipliedBy(i));
        int tickPos = (int) (viewWindow.getPixelPerMs() * (tickTime.getMillis() - viewWindow.getWindowStart().getMillis()));
        return new Tuple2<>(tickTime, tickPos);
    }

    protected static void iterateTicks(FontMetrics fontMetrics, TimeWindowData timeWindowData, WindowParams viewWindow, Period inbetweenTickSize, BiConsumer<Integer, Integer> inbetweenTickConsumer, BiConsumer<Boolean, Integer> tickConsumer, TriConsumer<Boolean, Integer, String> annotationConsumer) {
        DateTimeFormatter format = ToStrings.getFittingDateTimeFormatter(timeWindowData.getIntervalSize());

        Period intervalSize = timeWindowData.getIntervalSize();
        int from = timeWindowData.bin(viewWindow.getWindowStart());
        int to = timeWindowData.bin(viewWindow.getWindowEnd());

        int inbetweenTickCount = approximateInbetweenTickCount(viewWindow.getWindowSize(), inbetweenTickSize);
        boolean skipInbetween = false;
        if (inbetweenTickCount > 2000) skipInbetween = true; // will be more than one per pixel
        int tickCount = to - from + 1;
        int jumps = (int) Math.ceil(tickCount / 1000.0);

        Tuple2<DateTime, Integer> tuple2 = getInterval(viewWindow, intervalSize, from);
        DateTime time = tuple2._1;
        String print = format.print(time);
        int stringWidth = fontMetrics.stringWidth(print);

        int leftmostBelow = tuple2._2 - BIG_TICK_ANNOTATION_PADDING;
        int leftmostAbove = leftmostBelow + (stringWidth + BIG_TICK_ANNOTATION_PADDING) / 2;

        // guaranteed first tick on bottom
        Tuple2<Integer, Integer> above = new Tuple2<>(leftmostAbove, 0);
        Tuple2<Integer, Integer> below = new Tuple2<>(leftmostBelow, 0);

        boolean skipFirst = false;
        for (int i = from; i <= to; i += jumps) {
            if (inbetweenTickSize != null && inbetweenTickConsumer != null && !skipInbetween)
                skipFirst = iterateInbetweenTicks(viewWindow, intervalSize, inbetweenTickSize, i, skipFirst, inbetweenTickConsumer);
            iterateTick(fontMetrics, format, viewWindow, intervalSize, above, below, i, ((i - from % jumps) % (2 * jumps)) == 0, tickConsumer, annotationConsumer);
        }

    }

    private static boolean iterateInbetweenTicks(WindowParams viewWindow, Period intervalSize, Period inbetweenTickSize, int i, boolean skipFirst, BiConsumer<Integer, Integer> inbetweenTickConsumer) {
        if (inbetweenTickSize == null || inbetweenTickConsumer == null) {
            return false;
        }

        Tuple2<DateTime, Integer> fromTuple = getInterval(viewWindow, intervalSize, i);
        Tuple2<DateTime, Integer> toTuple = getInterval(viewWindow, intervalSize, i + 1);

        DateTime from = fromTuple._1;
        DateTime to = toTuple._1;
        if (skipFirst)
            from = from.plus(inbetweenTickSize);

        skipFirst = false;

        for (DateTime j = from; j.isBefore(to); j = j.plus(inbetweenTickSize.multipliedBy(2))) {
            int tickPos = (int) (viewWindow.getPixelPerMs() * (j.getMillis() - viewWindow.getWindowStart().getMillis()));
            DateTime min = j.plus(inbetweenTickSize);
            if (!min.isBefore(to)) {
                skipFirst = true;
                min = to;
            }
            int nextTickPos = (int) (viewWindow.getPixelPerMs() * (min.getMillis() - viewWindow.getWindowStart().getMillis()));

            inbetweenTickConsumer.accept(tickPos, nextTickPos);
        }

        return skipFirst;
    }

    private static void iterateTick(FontMetrics fontMetrics, DateTimeFormatter format, WindowParams viewWindow, Period intervalSize, Tuple2<Integer, Integer> above, Tuple2<Integer, Integer> below, int i, boolean isBelowTick, BiConsumer<Boolean, Integer> tickConsumer, TriConsumer<Boolean, Integer, String> annotationConsumer) {
        Tuple2<DateTime, Integer> tuple2 = getInterval(viewWindow, intervalSize, i);
        DateTime time = tuple2._1;
        int tickPos = tuple2._2;
        String print = format.print(time);
        int stringWidth = fontMetrics.stringWidth(print);

        int annotationPos = 0 <= tickPos && tickPos <= 2 ? 0 : tickPos - 2;
        if (isBelowTick) { // below
            if (tickConsumer != null)
                tickConsumer.accept(true, tickPos);
            if (tickPos - below._1 >= below._2 + BIG_TICK_ANNOTATION_PADDING) {
                if (annotationConsumer != null)
                    annotationConsumer.accept(true, annotationPos, print);
                below._1 = tickPos;
                below._2 = stringWidth;
            }
        } else { // above
            if (tickConsumer != null)
                tickConsumer.accept(false, tickPos);
            if (tickPos - above._1 >= above._2 + BIG_TICK_ANNOTATION_PADDING) {
                if (annotationConsumer != null)
                    annotationConsumer.accept(false, annotationPos, print);
                above._1 = tickPos;
                above._2 = stringWidth;
            }
        }
    }

    private static int approximateInbetweenTickCount(long duration, Period inbetweenTickSize) {
        if (inbetweenTickSize == null) return 0;
        double inbetweenLength = inbetweenTickSize.toDurationFrom(new Instant(0)).getMillis();
        double inbetweenTicks = duration / inbetweenLength;
        return ((int) inbetweenTicks);
    }

    public BufferedImage getImage() {
        BufferedImage image = new BufferedImage(getWidth() * 4, getHeight() * 4, BufferedImage.TYPE_INT_RGB);
        Graphics2D g2 = image.createGraphics();
        g2.setTransform(AffineTransform.getScaleInstance(4, 4));
        try {
            print(g2);
        } catch (Exception e) {
            e.printStackTrace();
        }
        return image;
    }

    protected void drawX(Graphics2D graphics2D) {
        graphics2D.setColor(Color.black);
        Stroke old = graphics2D.getStroke();
        graphics2D.setStroke(new BasicStroke(2));
        graphics2D.drawLine(0, 0, getWidth() - 1, getHeight() - 1);
        graphics2D.drawLine(0, getHeight() - 1, getWidth() - 1, 0);
        graphics2D.drawString("NO DATA", 5, 30);
        graphics2D.setStroke(old);
    }

    protected void drawBackground(Graphics2D graphics2D, PSMVisualSettings visualSettings, TimeWindowData timeWindowData, WindowParams viewWindow) {
        Color background = graphics2D.getBackground();
        Color alternative = accent;

        Color old = graphics2D.getColor();
        graphics2D.setColor(alternative);

        int height = getHeight();

        PSMVisualSettings.InbetweenTick inbetweenTick = visualSettings.getInbetweenTick();
        iterateTicks(graphics2D.getFontMetrics(), timeWindowData, viewWindow, inbetweenTick.getPeriod(), (tickPos, nextTickPos) -> {
            graphics2D.fillRect(tickPos, 0, nextTickPos - tickPos, height);
        }, null, null);

        graphics2D.setColor(old);
    }

    protected void drawHelpLines(Graphics2D graphics2D, TimeWindowData timeWindowData, WindowParams viewWindow) {
        graphics2D.setColor(Color.black);
        Stroke old = graphics2D.getStroke();
        BasicStroke ds = new BasicStroke(1.0f, BasicStroke.CAP_BUTT,
                BasicStroke.JOIN_MITER, 10.0f, new float[]{10.0f}, 5.0f);
        graphics2D.setStroke(ds);

        int height = getHeight();

        iterateTicks(graphics2D.getFontMetrics(), timeWindowData, viewWindow, null, null, (isBelow, tickPos) -> {
            graphics2D.drawLine(tickPos, 0, tickPos, height);
        }, null);

        graphics2D.setStroke(old);
    }

    protected void drawFrame(Graphics2D graphics2D, String title, int width, int height) {
        graphics2D.setStroke(new BasicStroke(2));
        graphics2D.setColor(Color.black);
        graphics2D.drawRect(0, 0, width - 1, height - 1);
        Font currentFont = graphics2D.getFont();
        Font newFont = currentFont.deriveFont(currentFont.getSize() * 1.3F);
        graphics2D.setFont(newFont);
        graphics2D.drawString(title, 5, 15);
    }


    @Override
    public void mouseClicked(MouseEvent e) {
    }

    protected JPopupMenu createPopupMenu(MouseEvent e) {
        JPopupMenu menu = null;
        MyDataSource dataSource = psmUpdateManager.getDataSource();
        if (dataSource.isComplete()) {
            int width = getWidth();
            WindowParams viewWindow = ViewWindowTools.getWindow(psmUpdateManager.getVisualSettings(), dataSource.getProcessStart(), dataSource.getProcessDuration(), width);
            double v = viewWindow.getWindowSize() * (double) e.getX() / width;
            DateTime time = viewWindow.getWindowStart().plus((long) v);

            menu = new JPopupMenu("current time");
            menu.add(new JMenuItem(DateTimeFormat.shortDateTime().print(time) + " (" + time.dayOfWeek().getAsShortText() + ")"));

            JMenuItem exportImageItem = new JMenuItem("export as image");
            exportImageItem.addActionListener(e1 -> {
                SwingWorker<BufferedImage, Void> sw = new SwingWorker<BufferedImage, Void>() {
                    @Override
                    protected BufferedImage doInBackground() throws Exception {
                        return getImage();
                    }

                    @Override
                    protected void done() {
                        try {
                            BufferedImage image = get();
                            JFileChooser fc = new JFileChooser();
                            fc.setAcceptAllFileFilterUsed(false);
                            fc.setFileFilter(new FileNameExtensionFilter("Image Files (jpg, png, tif)", "jpg", "png", "tif"));
                            fc.setSelectedFile(new File("vis.png"));
                            int userSelection = fc.showSaveDialog(null);
                            if (userSelection == JFileChooser.APPROVE_OPTION) {
                                File selectedFile = fc.getSelectedFile();
                                try {
                                    String name = selectedFile.getName();
                                    ImageIO.write(image, name.substring(name.lastIndexOf(".") + 1), selectedFile);
                                } catch (IOException ex) {
                                    JOptionPane.showMessageDialog(null, "Image could not be saved", "File Export Error", JOptionPane.ERROR_MESSAGE);
                                }
                            }
                        } catch (InterruptedException | ExecutionException ex) {
                            ex.printStackTrace();
                        }
                    }
                };
                sw.execute();
            });
            menu.add(exportImageItem);
        }
        return menu;
    }

    protected void showPopup(MouseEvent e) {
        SwingWorker<JPopupMenu, Void> sw = new SwingWorker<JPopupMenu, Void>() {
            @Override
            protected JPopupMenu doInBackground() throws Exception {
                return createPopupMenu(e);
            }

            @Override
            protected void done() {
                try {
                    JPopupMenu menu = get();
                    if (menu != null)
                        menu.show(e.getComponent(), e.getX(), e.getY());
                } catch (InterruptedException | ExecutionException ex) {
                    ex.printStackTrace();
                }
                super.done();
            }
        };

        sw.execute();
    }

    @Override
    public void mousePressed(MouseEvent e) {
        if (e.isPopupTrigger()) {
            showPopup(e);
        } else
            dragStart = e.getX();
    }

    @Override
    public void mouseReleased(MouseEvent e) {
        if (e.isPopupTrigger()) showPopup(e);
        else {
            dragStart = null;
            draggedPositive = null;
        }
    }

    @Override
    public void mouseEntered(MouseEvent e) {

    }

    @Override
    public void mouseExited(MouseEvent e) {

    }

    @Override
    public void mouseDragged(MouseEvent e) {
        PSMVisualSettings psmVisualSettings = psmUpdateManager.getVisualSettings();
        if (dragStart == null) return;
        int width = getWidth();

        int currentDragPos = e.getX();

        if (draggedPositive != null) {
            if ((draggedPositive && currentDragPos <= dragPosition) || (!draggedPositive && currentDragPos >= dragPosition)) {
                // direction change
                dragStart = dragPosition;
            }
        }

        int dx = dragStart - currentDragPos;
        double v = Math.abs(dx); // Math.sqrt(dx * dx + dy * dy);

        double rel = v / (width * psmVisualSettings.getXScale());

        double oldAnchor = psmVisualSettings.getViewAnchor();
        if (dx >= 0) { // dragged to the left
            psmVisualSettings.setViewAnchor(oldAnchor + rel);
            draggedPositive = true;
        } else {
            psmVisualSettings.setViewAnchor(oldAnchor - rel);
            draggedPositive = false;
        }
        dragPosition = currentDragPos;

        psmUpdateManager.visualSettingsChanged();
    }

    @Override
    public void mouseMoved(MouseEvent e) {

    }

    @Override
    public void mouseWheelMoved(MouseWheelEvent e) {
        PSMVisualSettings psmVisualSettings = psmUpdateManager.getVisualSettings();
        if (e.isShiftDown()) {
            double amount = e.getScrollAmount() * Math.signum(e.getWheelRotation());
            double xScale = psmVisualSettings.getXScale();
            double anchorDelta = amount / (20 * xScale);
            double oldAnchor = psmVisualSettings.getViewAnchor();
            psmVisualSettings.setViewAnchor(oldAnchor + anchorDelta);
        } else {
            double oldScale = psmVisualSettings.getXScale();

            // center view to where the mouse was scrolled
            double anchor = psmVisualSettings.getViewAnchor();
            double effectiveAnchor = effectiveAnchor(anchor, oldScale);

            int x = e.getX();
            int width = getWidth();
            double relMouse = (double) x / width - 0.5;

            double newAnchor = relMouse / (4 * oldScale) + effectiveAnchor;
            //

            double amount = Math.pow(1.1, e.getScrollAmount());
            double newScale;
            if (e.getWheelRotation() < 0)
                newScale = oldScale * amount;
            else
                newScale = oldScale / amount;

            psmVisualSettings.setXScale(newScale);
            psmVisualSettings.setViewAnchor(newAnchor);
        }
        psmUpdateManager.visualSettingsChanged();
    }

    private double effectiveAnchor(double anchor, double scale) {
        double res = anchor;
        double halfWindow = 1 / (scale * 2);
        if (anchor - halfWindow <= 0)
            res = halfWindow;
        else if (anchor + halfWindow >= 1)
            res = 1 - halfWindow;
        return res;
    }

    @Override
    public void updateDataSettings(DataSettings deltaSettings) {
        repaint();
    }


}
