package org.processmining.performancespectrumintegration.mypsm.segmentpanel;

import org.joda.time.DateTime;
import org.processmining.performancespectrumintegration.algorithms.segments.OrderedSegmentMeasurement;
import org.processmining.performancespectrumintegration.algorithms.segments.Segment;
import org.processmining.performancespectrumintegration.algorithms.segments.SegmentOccurrence;
import org.processmining.performancespectrumintegration.models.MyDataSource;
import org.processmining.performancespectrumintegration.mypsm.PSMColorMap;
import org.processmining.performancespectrumintegration.mypsm.ViewWindowTools;
import org.processmining.performancespectrumintegration.mypsm.update.MarkedCaseConnection;
import org.processmining.performancespectrumintegration.mypsm.update.PSMUpdateManager;
import org.processmining.performancespectrumintegration.mypsm.update.PSMVisualSettings;
import org.processmining.performancespectrumintegration.util.TriConsumer;
import org.processmining.performancespectrumintegration.util.Tuple2;

import javax.swing.*;
import java.awt.*;
import java.awt.event.MouseEvent;
import java.awt.geom.Line2D;

public class SegmentLinesPanel extends SegmentDataPanel {

    private final Segment segment;

    public SegmentLinesPanel(Segment segment, PSMUpdateManager psmUpdateManager) {
        super(psmUpdateManager);
        this.segment = segment;
    }

    @Override
    public void paintComponent(Graphics g) {
        super.paintComponent(g);

        int width = getWidth();
        int height = getHeight();

        Graphics2D graphics2D = (Graphics2D) g;

        graphics2D.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        graphics2D.clipRect(0, 0, width, height);

        MyDataSource dataSource = psmUpdateManager.getDataSource();
        PSMVisualSettings visualSettings = psmUpdateManager.getVisualSettings();

        OrderedSegmentMeasurement segmentMeasurement = dataSource.getSegmentMeasurement(segment);

        if (segmentMeasurement == null || segmentMeasurement.getMeasurementReliability().isUseless() || !dataSource.isComplete()) {
            drawX(graphics2D);
        } else {
            ViewWindowTools.WindowParams viewWindow = ViewWindowTools.getWindow(visualSettings, dataSource.getProcessStart(), dataSource.getProcessDuration(), width);

            if (visualSettings.isShowBackgroundPattern())
                drawBackground(graphics2D, visualSettings, dataSource.getTimeWindowData(), viewWindow);

            iterateLines(psmUpdateManager, viewWindow, (caseId, color, line) -> {
                graphics2D.setColor(color);
                graphics2D.draw(line);
            });

            if (visualSettings.isShowHelpLines())
                drawHelpLines(graphics2D, dataSource.getTimeWindowData(), viewWindow);
        }

        drawFrame(graphics2D, segment.toString() + " " + dataSource.getSegmentReliability(segment).toShortString(), width, height);

    }

    @Override
    protected JPopupMenu createPopupMenu(MouseEvent e) {
        JPopupMenu menu = super.createPopupMenu(e);
        MyDataSource dataSource = psmUpdateManager.getDataSource();
        MarkedCaseConnection markedCaseConnection = psmUpdateManager.getMarkedCaseConnection();
        if (menu != null && dataSource.isComplete()) {
            ViewWindowTools.WindowParams viewWindow = ViewWindowTools.getWindow(psmUpdateManager.getVisualSettings(), dataSource.getProcessStart(), dataSource.getProcessDuration(), getWidth());

            int x = e.getX();
            int y = e.getY();

            Tuple2<Integer, Double> closestClassId = new Tuple2<>(-1, Double.MAX_VALUE);
            iterateLines(psmUpdateManager, viewWindow, (caseId, color, line) -> {
                double v = line.ptLineDistSq(x, y);
                if (v < closestClassId._2) {
                    closestClassId._1 = caseId;
                    closestClassId._2 = v;
                }
            });
            JMenuItem markClosest = new JMenuItem("mark closest case id: " + closestClassId._1);
            markClosest.addActionListener(e1 -> {
                markedCaseConnection.setCaseId(closestClassId._1);
                markedCaseConnection.setMarked(true);
                psmUpdateManager.markedCaseConnectionChanged();
            });
            menu.add(markClosest);
            JMenuItem unmark = new JMenuItem("unmark case id");
            unmark.addActionListener(e1 -> {
                markedCaseConnection.setMarked(false);
                psmUpdateManager.markedCaseConnectionChanged();
            });
            menu.add(unmark);
            return menu;
        }
        return null;
    }

    private void iterateLines(PSMUpdateManager psmUpdateManager, ViewWindowTools.WindowParams viewWindow, TriConsumer<Integer, Color, Line2D.Float> consumer) {
        int height = getHeight();

        MyDataSource dataSource = psmUpdateManager.getDataSource();
        PSMVisualSettings visualSettings = psmUpdateManager.getVisualSettings();
        PSMColorMap psmColorMap = psmUpdateManager.getPsmColorMap();
        MarkedCaseConnection markedCaseConnection = psmUpdateManager.getMarkedCaseConnection();
        OrderedSegmentMeasurement segmentMeasurement = dataSource.getSegmentMeasurement(segment);

        Iterable<SegmentOccurrence> subset;
        if (visualSettings.isSubsetApproximation())
            subset = segmentMeasurement.subsetApproximation(new DateTime(viewWindow.getWindowStart()), new DateTime(viewWindow.getWindowEnd()));
        else
            subset = segmentMeasurement.subset(new DateTime(viewWindow.getWindowStart()), new DateTime(viewWindow.getWindowEnd()));

        int reduction = visualSettings.getLineDensity().getReduction();
        int middle = reduction / 2;

        int i = 0;
        for (SegmentOccurrence so : subset) {
            Integer caseId = so.getCaseId();
            Integer classId = dataSource.getClass(caseId);

            if (!psmColorMap.isActive(classId))
                continue;

            if (i++ % reduction != middle) continue;

            Color color;
            if (markedCaseConnection.isMarked() && markedCaseConnection.getCaseId() == caseId)
                color = Color.green;
            else
                color = psmColorMap.getColor(classId);

            double relStart = so.getStartTime().getMillis() - viewWindow.getWindowStart().getMillis();
            double length = so.getDuration().getMillis();

            double scaledStart = relStart * viewWindow.getPixelPerMs();
            double scaledLength = length * viewWindow.getPixelPerMs();

            int x1 = (int) scaledStart;
            int y1 = 0;
            int x2 = (int) (scaledStart + scaledLength);
            int y2 = height;

            consumer.accept(caseId, color, new Line2D.Float(x1, y1, x2, y2));
        }
    }


}
