package org.processmining.performancespectrumintegration.mypsm.update;

import org.joda.time.Duration;
import org.joda.time.Instant;
import org.joda.time.Period;
import org.processmining.performancespectrumintegration.util.Tuple2;

import java.util.Arrays;

public class PSMVisualSettings {

    public static final double MIN_YSCALE = .62;
    public static final int MAX_YSCALE = 4;
    public static final int MAX_XSCALE = 1000;
    private boolean showHelpLines;
    private double yScale, viewAnchor;
    private InbetweenTick inbetweenTick;
    private boolean subsetApproximation;
    private LineDensity lineDensity;

    private boolean showBackgroundPattern;
    private ExponentialScale expScale;

    public PSMVisualSettings() {
        reset();
    }

    public boolean isShowHelpLines() {
        return showHelpLines;
    }

    public void setShowHelpLines(boolean showHelpLines) {
        this.showHelpLines = showHelpLines;
    }

    public void reset() {
        yScale = 1;
        viewAnchor = 0.5;
        showHelpLines = false;
        showBackgroundPattern = false;
        inbetweenTick = InbetweenTick.Week;
        subsetApproximation = false;
        lineDensity = LineDensity.Full;
        expScale = new ExponentialScale(1);
    }

    public double getXScale() {
        return expScale.getEffectiveScale();
    }

    public void setXScale(double xScale) {
        expScale.setEffectiveScale(xScale);
    }

    @Override
    public String toString() {
        return "PSMVisualSettings{" +
                "showHelpLines=" + showHelpLines +
                ", xScale=" + expScale.getEffectiveScale() +
                ", yScale=" + yScale +
                ", viewAnchor=" + viewAnchor +
                ", inbetweenTick=" + inbetweenTick +
                ", subsetApproximation=" + subsetApproximation +
                ", linesReduction=" + lineDensity.getReduction() +
                ", showBackgroundPattern=" + showBackgroundPattern +
                '}';
    }

    public double getViewAnchor() {
        return viewAnchor;
    }

    public void setViewAnchor(double viewAnchor) {
        if (viewAnchor >= 1)
            this.viewAnchor = 1;
        else if (viewAnchor <= 0 || Double.isNaN(viewAnchor))
            this.viewAnchor = 0;
        else
            this.viewAnchor = viewAnchor;
    }

    public boolean isSubsetApproximation() {
        return subsetApproximation;
    }

    public void setSubsetApproximation(boolean subsetApproximation) {
        this.subsetApproximation = subsetApproximation;
    }

    public InbetweenTick getInbetweenTick() {
        return inbetweenTick;
    }

    public void setInbetweenTick(InbetweenTick duration) {
        inbetweenTick = duration;
    }

    public boolean isShowBackgroundPattern() {
        return showBackgroundPattern;
    }

    public void setShowBackgroundPattern(boolean showBackgroundPattern) {
        this.showBackgroundPattern = showBackgroundPattern;
    }

    public double getYScale() {
        return yScale;
    }

    public void setYScale(double yScale) {
        if (yScale >= MAX_YSCALE)
            this.yScale = MAX_YSCALE;
        else if (yScale <= MIN_YSCALE)
            this.yScale = MIN_YSCALE;
        else
            this.yScale = yScale;
    }

    public ExponentialScale getExponentialScale() {
        return expScale;
    }

    public LineDensity getLineDensity() {
        return lineDensity;
    }

    public void setLineDensity(LineDensity lineDensity) {
        this.lineDensity = lineDensity;
    }

    public enum InbetweenTick {
        Year(Period.years(1)), Quarter(Period.months(3)),
        Month(Period.months(1)), Week(Period.weeks(1)),
        Day(Period.days(1)), Hour(Period.hours(1)),
        Minute(Period.minutes(1)), Second(Period.seconds(1));

        private final Period period;

        InbetweenTick(Period period) {
            this.period = period;
        }

        public static InbetweenTick[] valuesShorterThan(Period period) {
            Duration d = period.toDurationFrom(new Instant(0));
            InbetweenTick[] arr = Arrays.stream(values()).filter(it -> it.getPeriod().toDurationFrom(new Instant(0)).isShorterThan(d)).toArray(InbetweenTick[]::new);
            if (arr.length == 0)
                return new InbetweenTick[]{Second};
            else
                return arr;
        }

        public Period getPeriod() {
            return period;
        }
    }

    public enum LineDensity {
        Full("100%", 1), Half("50%", 2), Third("33%", 3), Fifth("20%", 5), Tenth("10%", 10), Twentieth("5%", 20);

        private final String name;
        private final int reduction;

        LineDensity(String name, int reduction) {
            this.name = name;
            this.reduction = reduction;
        }

        @Override
        public String toString() {
            return name;
        }

        public int getReduction() {
            return reduction;
        }
    }

    public static class ExponentialScale {

        // log_BASE(MAX_EFF) = MAX_EXP
        private static final double MAX_EXP = 200d;
        private static final double BASE = 1.047128548;
        private static final double MAX_EFF = Math.pow(BASE, MAX_EXP); // 10000
        private double exponent = 0;

        public ExponentialScale(double scale) {
            setEffectiveScale(scale);
        }

        public Tuple2<Double, Double> getExponentRange() {
            return new Tuple2<>(0d, MAX_EXP);
        }

        public Tuple2<Double, Double> getEffectiveRange() {
            return new Tuple2<>(1d, MAX_EFF);
        }

        public double getEffectiveScale() {
            return Math.pow(BASE, exponent);
        }

        public void setEffectiveScale(double scale) {
            if (scale >= MAX_EFF)
                exponent = MAX_EXP;
            else if (scale <= 1) exponent = 0;
            else
                exponent = Math.log(scale) / Math.log(BASE);
        }

        public double getExponent() {
            return exponent;
        }

        public void setExponent(double exponent) {
            if (exponent >= MAX_EXP)
                this.exponent = MAX_EXP;
            else if (exponent <= 0)
                this.exponent = 0;
            else
                this.exponent = exponent;
        }

    }

}
