package org.processmining.performancespectrumintegration.parameters.serialization;

import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.DeserializationContext;
import com.fasterxml.jackson.databind.KeyDeserializer;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.common.collect.Sets;
import org.processmining.models.graphbased.LocalNodeID;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.performancespectrumintegration.parameters.LocalNodeIDMixIn;
import org.processmining.performancespectrumintegration.parameters.PSMParameters;
import org.processmining.performancespectrumintegration.util.Tuple2;
import org.processmining.performancespectrumintegration.visualization.PetrinetCopyConnection;
import org.processmining.performancespectrumintegration.visualization.PlaceSelectionMap;
import org.processmining.performancespectrumintegration.visualization.PlaceSelectionState;
import org.processmining.performancespectrumintegration.visualization.controller.SegmentMeasurementController;

import java.io.IOException;
import java.io.Serializable;
import java.util.Objects;
import java.util.Set;
import java.util.stream.Collectors;

public class PrimitivePlaceSelectionMap implements Serializable {

    @JsonProperty
    //@JsonDeserialize(keyUsing = MapKeyDeserializer.class)
    private Set<Tuple2<LocalNodeID, PlaceSelectionState>> places;
    @JsonProperty
    private Set<Tuple2<Tuple2<Set<LocalNodeID>, Set<LocalNodeID>>, PlaceSelectionState>> measurementPlaces;

    private PrimitivePlaceSelectionMap() {
    }

    public PrimitivePlaceSelectionMap(PlaceSelectionMap placeSelectionMap) {
        places = Sets.newHashSet();
        for (Tuple2<Place, PlaceSelectionState> tuple2 : placeSelectionMap.originalIterable()) {
            if (tuple2._2 != PlaceSelectionState.Off)
                places.add(new Tuple2<>(tuple2._1.getLocalID(), tuple2._2));
        }
        measurementPlaces = Sets.newHashSet();
        for (Tuple2<Place, PlaceSelectionState> tuple2 : placeSelectionMap.measurementIterable()) {
            measurementPlaces.add(new Tuple2<>(getAdjacentLocalIds(tuple2._1, placeSelectionMap.getPetrinetConnection(), placeSelectionMap.getPetrinet()), tuple2._2));
        }
    }

    private Tuple2<Set<LocalNodeID>, Set<LocalNodeID>> getAdjacentLocalIds(Place place, PetrinetCopyConnection copyConnection, Petrinet net) {
        Set<LocalNodeID> collectIn = net.getInEdges(place).stream().map(e -> copyConnection.getOriginalTransition((Transition) e.getSource()).getLocalID()).collect(Collectors.toSet());
        Set<LocalNodeID> collectOut = net.getOutEdges(place).stream().map(e -> copyConnection.getOriginalTransition((Transition) e.getTarget()).getLocalID()).collect(Collectors.toSet());
        return new Tuple2<>(collectIn, collectOut);
    }

    public void recreate(PlaceSelectionMap placeSelectionMap, SegmentMeasurementController measurementController, Petrinet net) {
        boolean notFound = false;
        PetrinetCopyConnection copyConnection = placeSelectionMap.getPetrinetConnection();
        for (Tuple2<LocalNodeID, PlaceSelectionState> tuple2 : places) {
            Place place = PSMParameters.findPlace(tuple2._1, net);
            if (place != null) {
                if (net.getInEdges(place).size() == 1 && net.getOutEdges(place).size() == 1)
                    placeSelectionMap.setState(copyConnection.getCopyPlace(place), PlaceSelectionState.Individual);
                else if (net.getInEdges(place).size() > 1 || net.getOutEdges(place).size() > 1)
                    placeSelectionMap.setState(copyConnection.getCopyPlace(place), tuple2._2); // should be done before listener is added
                // ignore places that wouldn't lead to segments
            } else {
                notFound = true;
                System.out.println("No corresponding place in the petrinet could be found for " + tuple2._1 + " (" + tuple2._2 + ")");
            }
        }

        if (notFound)
            System.out.println(net.getPlaces().stream().map(PetrinetNode::getLocalID).sorted().collect(Collectors.toList()));

        for (Tuple2<Tuple2<Set<LocalNodeID>, Set<LocalNodeID>>, PlaceSelectionState> tuple2 : measurementPlaces) {
            PlaceSelectionState state = tuple2._2;
            Set<Transition> preset = tuple2._1._1.stream().map(id -> PSMParameters.findTransition(id, net)).filter(Objects::nonNull).map(copyConnection::getCopyTransition).collect(Collectors.toSet());
            Set<Transition> postset = tuple2._1._2.stream().map(id -> PSMParameters.findTransition(id, net)).filter(Objects::nonNull).map(copyConnection::getCopyTransition).collect(Collectors.toSet());
            measurementController.setSourceTransitions(preset);
            measurementController.setTargetTransitions(postset);
            Place place = measurementController.addMeasurementPlace(false, true);
            placeSelectionMap.setState(place, state);
        }
    }

    public static class MapKeyDeserializer extends KeyDeserializer {

        private static final ObjectMapper mapper = new ObjectMapper();

        static {
            mapper.addMixIn(LocalNodeID.class, LocalNodeIDMixIn.class);
        }

        @Override
        public Object deserializeKey(String key, DeserializationContext ctxt) throws IOException, JsonProcessingException {
            return mapper.readValue(key, LocalNodeID.class);
        }
    }

}
