package org.processmining.performancespectrumintegration.plugins;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.Sets;
import nl.tue.astar.AStarException;
import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.model.XLog;
import org.processmining.acceptingpetrinet.models.AcceptingPetriNet;
import org.processmining.acceptingpetrinet.models.impl.AcceptingPetriNetFactory;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.*;
import org.processmining.log.utils.XUtils;
import org.processmining.models.connections.petrinets.behavioral.FinalMarkingConnection;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.graphbased.AttributeMap;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.performancespectrumintegration.models.InteractivePSM;
import org.processmining.performancespectrumintegration.parameters.PSMParameters;
import org.processmining.plugins.astar.petrinet.PetrinetReplayerWithILP;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;
import org.processmining.plugins.petrinet.replayer.PNLogReplayer;
import org.processmining.plugins.petrinet.replayer.algorithms.IPNReplayParameter;
import org.processmining.plugins.petrinet.replayer.algorithms.costbasedcomplete.CostBasedCompleteParam;
import org.processmining.plugins.petrinet.replayresult.PNRepResult;
import org.processmining.pnetreplayer.utils.TransEvClassMappingUtils;

import javax.swing.*;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

@Plugin(name = "Interactive Performance Spectrum", parameterLabels = {"Event Log", "Petrinet", "Psm Parameters"}, level = PluginLevel.PeerReviewed, quality = PluginQuality.Good, icon = "icon.png",
        returnLabels = {"Interactive PSM"}, returnTypes = {InteractivePSM.class}, help = InteractivePSMPlugin.HELP, categories = {PluginCategory.Analytics}, keywords = {"performance", "interactive", "psm"})
public class InteractivePSMPlugin {

    public static final String HELP = "This plugin provides an interactive integration of the PSM with Petrinets and Alignments.";

    @UITopiaVariant(affiliation = "PADS RWTH", author = "Daniel Tacke genannt Unterberg", email = "daniel.tacke.unterberg@rwth-aachen.de")
    @PluginVariant(variantLabel = "Interactive PSM", requiredParameterLabels = {0, 1})
    public InteractivePSM run(UIPluginContext context, XLog log, Petrinet petrinet) {
        return runPrivate(context, log, petrinet, null);
    }

    @UITopiaVariant(affiliation = "PADS RWTH", author = "Daniel Tacke genannt Unterberg", email = "daniel.tacke.unterberg@rwth-aachen.de")
    @PluginVariant(variantLabel = "Interactive PSM", requiredParameterLabels = {0, 1, 2})
    public InteractivePSM run(UIPluginContext context, XLog log, Petrinet petrinet, PSMParameters psmParameters) {
        return runPrivate(context, log, petrinet, psmParameters);
    }

    @UITopiaVariant(affiliation = "PADS RWTH", author = "Daniel Tacke genannt Unterberg", email = "daniel.tacke.unterberg@rwth-aachen.de")
    @PluginVariant(variantLabel = "Interactive PSM", requiredParameterLabels = {0, 1})
    public InteractivePSM runAccepting(UIPluginContext context, XLog log, AcceptingPetriNet acceptingPetriNet) {
        unpackAcceptingForAlignments(context, acceptingPetriNet);
        return runPrivate(context, log, acceptingPetriNet, null);
    }

    @UITopiaVariant(affiliation = "PADS RWTH", author = "Daniel Tacke genannt Unterberg", email = "daniel.tacke.unterberg@rwth-aachen.de")
    @PluginVariant(variantLabel = "Interactive PSM", requiredParameterLabels = {0, 1, 2})
    public InteractivePSM runAccepting(UIPluginContext context, XLog log, AcceptingPetriNet acceptingPetriNet, PSMParameters psmParameters) {
        unpackAcceptingForAlignments(context, acceptingPetriNet);
        return runPrivate(context, log, acceptingPetriNet, psmParameters);
    }

    /*
    @UITopiaVariant(affiliation = "PADS RWTH", author = "Daniel Tacke genannt Unterberg", email = "daniel.tacke.unterberg@rwth-aachen.de")
    @PluginVariant(variantLabel = "Interactive PSM (quick)", requiredParameterLabels = {0, 1})
    */
    public InteractivePSM runQuick(UIPluginContext context, XLog log, AcceptingPetriNet accepting) {
        Petrinet petrinet = accepting.getNet();

        XEventClassifier defaultClassifier = XUtils.getDefaultClassifier(log);

        XEventClasses xEventClasses = XEventClasses.deriveEventClasses(defaultClassifier, log);
        HashSet<XEventClass> classes = Sets.newHashSet(xEventClasses.getClasses());
        TransEvClassMapping mapping = TransEvClassMappingUtils.getInstance().getMapping(petrinet, classes, defaultClassifier);

        IPNReplayParameter param = new CostBasedCompleteParam(classes, mapping.getDummyEventClass(), petrinet.getTransitions(), 1, 1);

        //AcceptingPetriNet accepting = createAccepting(context, petrinet);

        param.setInitialMarking(accepting.getInitialMarking());
        param.setFinalMarkings(accepting.getFinalMarkings().toArray(new Marking[0]));
        param.setGUIMode(false);

        PetrinetReplayerWithILP replayer = new PetrinetReplayerWithILP();

        PNRepResult pnRepResult = null;

        try {
            pnRepResult = replayer.replayLog(context, petrinet, log, mapping, param);
        } catch (AStarException e) {
            e.printStackTrace();
        }

        if (pnRepResult == null) {
            context.getFutureResult(0).cancel(true);
            return null;
        } else return runPrivate(context, log, accepting, pnRepResult, null);
    }

    private void unpackAcceptingForAlignments(UIPluginContext context, AcceptingPetriNet acceptingPetriNet) {
        context.getConnectionManager().addConnection(new InitialMarkingConnection(acceptingPetriNet.getNet(), acceptingPetriNet.getInitialMarking()));
        context.getConnectionManager().addConnection(new FinalMarkingConnection(acceptingPetriNet.getNet(), acceptingPetriNet.getFinalMarkings().toArray(new Marking[0])[0]));
    }

    private InteractivePSM runPrivate(UIPluginContext context, XLog log, Petrinet petrinet, PSMParameters psmParameters) {
        AcceptingPetriNet acceptingPetriNet = createAccepting(context, petrinet);

        return runPrivate(context, log, acceptingPetriNet, psmParameters);
    }

    private InteractivePSM runPrivate(UIPluginContext context, XLog log, AcceptingPetriNet acceptingPetriNet, PSMParameters psmParameters) {
        PNLogReplayer replayer = new PNLogReplayer();

        PNRepResult result = null;

        try {
            result = replayer.replayLogGUI(context, acceptingPetriNet.getNet(), log);
        } catch (final ConnectionCannotBeObtained | AStarException connectionCannotBeObtained) {
            context.log("Could not create alignment");
            connectionCannotBeObtained.printStackTrace();
        }

        if (result == null) {
            JOptionPane.showMessageDialog(null, "alignment creation failed", "Alignment computation", JOptionPane.WARNING_MESSAGE);
            context.getFutureResult(0).cancel(true);
            return null;
        } else return runPrivate(context, log, acceptingPetriNet, result, psmParameters);
    }

    private InteractivePSM runPrivate(UIPluginContext context, XLog log, AcceptingPetriNet acceptingPetriNet, PNRepResult pnRepResult, PSMParameters psmParameters) {
        context.getFutureResult(0).setLabel("Interactive PSM on " + acceptingPetriNet.getNet().getLabel());

        System.out.println(pnRepResult.getInfo());

        Double fitness = (Double) pnRepResult.getInfo().get(PNRepResult.TRACEFITNESS);
        if (fitness < 0.7) {
            JOptionPane.showMessageDialog(null, "fitness is below 0.7, the visualizations might be misleading", "Fitness Warning", JOptionPane.WARNING_MESSAGE);
        } else if (fitness < 0.4) {
            // TODO fix and give better message
            context.log("Alignment has fitness of less than 0.4, continuing is pointless");
            //context.showConfiguration("Execution Aborted", new JLabel("fitness is below 0.4, continuing is pointless"));
            JOptionPane.showMessageDialog(null, "fitness is below 0.4, continuing is pointless", "Fitness Warning", JOptionPane.ERROR_MESSAGE);
            context.getFutureResult(0).cancel(true);
            return null;
        }

        return new InteractivePSM(log, acceptingPetriNet, pnRepResult, psmParameters);
    }

    private AcceptingPetriNet createAccepting(PluginContext context, Petrinet net) {
        Marking initialMarking;
        Set<Marking> finalMarkings;
        try {
            initialMarking = context.tryToFindOrConstructFirstObject(Marking.class, InitialMarkingConnection.class,
                    InitialMarkingConnection.MARKING, net);
        } catch (ConnectionCannotBeObtained e) {
            initialMarking = new Marking();
            for (Place place : net.getPlaces()) {
                if (net.getInEdges(place).isEmpty()) {
                    initialMarking.add(place);
                }
            }
        }
        try {
            Collection<Marking> finalMarkingCollection = context.tryToFindOrConstructAllObjects(Marking.class,
                    FinalMarkingConnection.class, FinalMarkingConnection.MARKING, net);
            if (finalMarkingCollection == null)
                System.out.println("final markings were null..");
            finalMarkings = new HashSet<>(finalMarkingCollection);
        } catch (ConnectionCannotBeObtained e) {
            finalMarkings = new HashSet<>();
            for (Place place : net.getPlaces()) {
                if (net.getOutEdges(place).isEmpty()) {
                    Marking finalMarking = new Marking();
                    finalMarking.add(place);
                    finalMarkings.add(finalMarking);
                }
            }
            if (finalMarkings.isEmpty()) {
                finalMarkings.add(new Marking());
            }
        }
        return AcceptingPetriNetFactory.createAcceptingPetriNet(net, initialMarking, finalMarkings);
    }

    private void makeNetUnique(Petrinet net) {
        HashMultimap<String, Transition> labelMap = HashMultimap.create();

        for (Transition t : net.getTransitions()) labelMap.put(t.getLabel(), t);

        labelMap.asMap().forEach((key, value) -> {
            int i = 1;
            for (Transition t : value) {
                if (t.isInvisible()) {
                    t.getAttributeMap().put(AttributeMap.LABEL, key + (i > 1 ? i : ""));
                    i++;
                }
            }
        });
    }

}
