package org.processmining.performancespectrumintegration.visualization;

import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import org.processmining.models.graphbased.AttributeMap;
import org.processmining.models.graphbased.ViewSpecificAttributeMap;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.performancespectrumintegration.algorithms.petrinet.GraphTools;
import org.processmining.performancespectrumintegration.algorithms.segments.MultiSegment;
import org.processmining.performancespectrumintegration.algorithms.segments.Segment;
import org.processmining.performancespectrumintegration.algorithms.segments.SingleSegment;
import org.processmining.performancespectrumintegration.mypsm.update.PSMUpdateManager;
import org.processmining.performancespectrumintegration.mypsm.update.PlaceSegmentConnection;
import org.processmining.performancespectrumintegration.util.Tuple2;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import java.util.*;
import java.util.stream.Collectors;

public class PlaceSelectionMap {

    private final Map<Place, PlaceSelectionState> internal;
    private final Petrinet petrinet;
    private final PSMUpdateManager psmUpdateManager;
    private final PetrinetCopyConnection petrinetConnection;
    private ChangeListener changeListener;

    public PlaceSelectionMap(Petrinet petrinet, PSMUpdateManager psmUpdateManager, PetrinetCopyConnection petrinetConnection) {
        this.petrinet = petrinet;
        this.psmUpdateManager = psmUpdateManager;
        this.petrinetConnection = petrinetConnection;
        this.internal = Maps.newHashMap();
        for (Place place : petrinet.getPlaces()) {
            addState(place);
        }
    }

    public Petrinet getPetrinet() {
        return petrinet;
    }

    public PetrinetCopyConnection getPetrinetConnection() {
        return petrinetConnection;
    }

    public void setChangeListener(ChangeListener changeListener) {
        this.changeListener = changeListener;
    }

    public PlaceSelectionState addState(Place place) {
        setState(place, PlaceSelectionState.getDefault());
        return PlaceSelectionState.getDefault();
    }

    public void addState(Place place, ViewSpecificAttributeMap map) {
        PlaceSelectionState state = addState(place);
        setPlaceAttributes(place, state, map);
    }

    public void setState(Place place, PlaceSelectionState state) {
        synchronized (internal) {
            internal.put(place, state);
            if (changeListener != null) changeListener.stateChanged(new ChangeEvent(this));
        }
    }

    public void setState(Place place, PlaceSelectionState state, ViewSpecificAttributeMap map) {
        setState(place, state);
        setPlaceAttributes(place, state, map);
    }

    public void updateAttributes(ViewSpecificAttributeMap map) {
        for (Map.Entry<Place, PlaceSelectionState> entry : internal.entrySet()) {
            setPlaceAttributes(entry.getKey(), entry.getValue(), map);
        }
    }

    public PlaceSelectionState toggleState(Place place) {
        PlaceSelectionState next = internal.get(place).next();
        setState(place, next);
        return next;
    }

    public PlaceSelectionState toggleState(Place place, ViewSpecificAttributeMap map) {
        PlaceSelectionState next = toggleState(place);
        setPlaceAttributes(place, next, map);
        return next;
    }

    public void removeState(Place place) {
        synchronized (internal) {
            internal.remove(place);
            if (changeListener != null) changeListener.stateChanged(new ChangeEvent(this));
        }
    }

    public void removeStates(Collection<Place> places) {
        synchronized (internal) {
            places.forEach(internal::remove);
            if (changeListener != null) changeListener.stateChanged(new ChangeEvent(this));
        }
    }

    public void removeState(Place place, ViewSpecificAttributeMap map) {
        removeState(place);
        removePlaceAttributes(place, map);
    }

    public void removeStates(Collection<Place> places, ViewSpecificAttributeMap map) {
        removeStates(places);
        places.forEach(p -> removePlaceAttributes(p, map));
    }

    public PlaceSelectionState getState(Place place) {
        return internal.get(place);
    }

    private void removePlaceAttributes(Place place, ViewSpecificAttributeMap map) {
        map.removeViewSpecific(place, AttributeMap.TOOLTIP);
        map.removeViewSpecific(place, AttributeMap.FILLCOLOR);
    }

    private void setPlaceAttributes(Place place, PlaceSelectionState state, ViewSpecificAttributeMap map) {
        map.putViewSpecific(place, AttributeMap.TOOLTIP, state.toString());
        map.putViewSpecific(place, AttributeMap.FILLCOLOR, state.getColor());
    }

    public Iterable<Tuple2<Place, PlaceSelectionState>> originalIterable() {
        return new Iterable<Tuple2<Place, PlaceSelectionState>>() {
            @Override
            public Iterator<Tuple2<Place, PlaceSelectionState>> iterator() {
                return internal.entrySet().stream().map(e -> {
                    if (petrinetConnection.hasOriginalPlace(e.getKey()))
                        return new Tuple2<>(petrinetConnection.getOriginalPlace(e.getKey()), e.getValue());
                    else return null;
                }).filter(Objects::nonNull).iterator();
            }
        };
    }

    public Iterable<Tuple2<Place, PlaceSelectionState>> measurementIterable() {
        return new Iterable<Tuple2<Place, PlaceSelectionState>>() {
            @Override
            public Iterator<Tuple2<Place, PlaceSelectionState>> iterator() {
                return internal.entrySet().stream().map(e -> {
                    if (petrinetConnection.hasOriginalPlace(e.getKey())) return null;
                    else return new Tuple2<>(e.getKey(), e.getValue());
                }).filter(Objects::nonNull).iterator();
            }
        };
    }

    public Set<Segment> computeSegments() {
        synchronized (internal) {
            PlaceSegmentConnection conn = psmUpdateManager.getPlaceSegmentConnection();
            conn.reset();
            Set<Segment> result = Sets.newHashSet();
            for (Map.Entry<Place, PlaceSelectionState> entry : internal.entrySet()) {
                switch (entry.getValue()) {
                    case Off:
                        break;
                    case Individual:
                        for (Transition t1 : GraphTools.getIngoingTransitions(petrinet, entry.getKey())) {
                            Transition from = petrinetConnection.getOriginalTransition(t1);
                            for (Transition t2 : GraphTools.getOutgoingTransitions(petrinet, entry.getKey())) {
                                Transition to = petrinetConnection.getOriginalTransition(t2);
                                SingleSegment ss = new SingleSegment(from, to);
                                result.add(ss);
                                conn.add(entry.getKey(), ss);
                            }
                        }
                        break;
                    case Aggregate:
                        Set<Transition> from = GraphTools.getIngoingTransitions(petrinet, entry.getKey()).stream().map(petrinetConnection::getOriginalTransition).collect(Collectors.toSet());
                        Set<Transition> to = GraphTools.getOutgoingTransitions(petrinet, entry.getKey()).stream().map(petrinetConnection::getOriginalTransition).collect(Collectors.toSet());
                        if (!from.isEmpty() && !to.isEmpty()) {
                            MultiSegment ms = new MultiSegment(from, to);
                            result.add(ms);
                            conn.add(entry.getKey(), ms);
                        }
                        break;
                }
            }
            conn.applyHighlighting();
            return result;
        }
    }

    public Set<Segment> computeAllPossibleSingleSegments() {
        Set<Segment> collect = internal.keySet().stream().flatMap(p ->
                GraphTools.getIngoingTransitions(petrinet, p).stream().map(petrinetConnection::getOriginalTransition).flatMap(t1 -> GraphTools.getOutgoingTransitions(petrinet, p).stream().map(petrinetConnection::getOriginalTransition).map(t2 -> new SingleSegment(t1, t2)))).collect(Collectors.toSet());
        return collect;
    }

    public Set<Segment> computeAllPossibleAggregateSegments() {
        Set<Segment> result = Sets.newHashSet();
        for (Place p : internal.keySet()) {
            Set<Transition> from = GraphTools.getIngoingTransitions(petrinet, p).stream().map(petrinetConnection::getOriginalTransition).collect(Collectors.toSet());
            Set<Transition> to = GraphTools.getOutgoingTransitions(petrinet, p).stream().map(petrinetConnection::getOriginalTransition).collect(Collectors.toSet());
            if (!from.isEmpty() && !to.isEmpty())
                result.add(new MultiSegment(from, to));
        }
        return result;
    }

    public Collection<Place> resetAll(ViewSpecificAttributeMap map) {
        synchronized (internal) {
            for (Place place : petrinet.getPlaces()) {
                internal.put(place, PlaceSelectionState.getDefault());
                setPlaceAttributes(place, PlaceSelectionState.getDefault(), map);
            }
            if (changeListener != null) changeListener.stateChanged(new ChangeEvent(this));
        }
        return petrinet.getPlaces();
    }

}
