package org.processmining.performancespectrumintegration.visualization;

import com.fluxicon.slickerbox.factory.SlickerFactory;
import org.joda.time.DateTime;
import org.joda.time.Duration;
import org.joda.time.Period;
import org.joda.time.format.DateTimeFormat;
import org.joda.time.format.PeriodFormat;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.performancespectrumintegration.algorithms.segments.Segment;
import org.processmining.performancespectrumintegration.models.InteractivePSM;
import org.processmining.performancespectrumintegration.models.settings.*;
import org.processmining.performancespectrumintegration.mypsm.ClassificationLegendFrame;
import org.processmining.performancespectrumintegration.mypsm.update.PSMAggregationSettings;
import org.processmining.performancespectrumintegration.mypsm.update.PSMUpdateManager;
import org.processmining.performancespectrumintegration.parameters.PSMParameters;
import org.processmining.performancespectrumintegration.util.ToStrings;
import org.processmining.performancespectrumintegration.util.Tuple2;
import org.processmining.performancespectrumintegration.util.Utils;
import org.processmining.performancespectrumintegration.util.WrapLayout;
import org.processmining.performancespectrumintegration.visualization.controller.ExecutionStatusConnection;
import org.processmining.performancespectrumintegration.visualization.controller.ExecutionStatusListener;
import org.processmining.performancespectrumintegration.visualization.controller.SettingsController;
import org.processmining.performancespectrumintegration.visualization.inputpanels.DialogPanel;
import org.processmining.plugins.utils.ProvidedObjectHelper;

import javax.swing.*;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutionException;

public class SettingsPanel extends JPanel implements ExecutionStatusListener {

    public static final ImageIcon gearIcon = Utils.loadImageIcon("gear_s.png");//new ImageIcon("lib/images/gear_s.png", "there are unapplied data changes not reflected in the visualization");
    public static final ImageIcon checkmarkIcon = Utils.loadImageIcon("checkmark_s.png");//new ImageIcon("lib/images/checkmark_s.png");
    public static final ImageIcon loadingIcon = Utils.loadImageIcon("loading_s.png");//new ImageIcon("lib/images/loading_s.png", "settings are being applied");

    private final PSMUpdateManager psmUpdateManager;
    private final InteractivePSM interactivePSM;
    private final UIPluginContext context;
    private JCheckBox instantModeCheckBox;
    private SettingsController settingsController;
    private JComboBox aggregationComboBox;
    private JLabel unappliedChangesLabel;
    private JButton applySettingsButton;
    private JCheckBox lockSegmentOrderCheckBox;


    public SettingsPanel(UIPluginContext context, TopPanel topPanel, InteractivePSM interactivePSM) {
        super(new WrapLayout());
        this.context = context;
        this.interactivePSM = interactivePSM;

        //setPreferredSize(new Dimension(1200, 75));
        //setSize(new Dimension(1200, 75));

        settingsController = topPanel.getSettingsController();
        psmUpdateManager = topPanel.getPSMUpdateManager();

        setupApplySettings();

        setupFilterSettings();

        setupClassificationSettings();

        setupAggregationSettings();

        JButton splitScreenButton = SlickerFactory.instance().createButton("toggle splitscreen");
        splitScreenButton.addActionListener(e -> SwingUtilities.invokeLater(topPanel::toggleSplitMode));

        add(splitScreenButton);

        JButton exportButton = SlickerFactory.instance().createButton("export");
        exportButton.addActionListener(e -> {
            if (context == null) return;

            PSMParameters psmParameters = settingsController.getPSMParameters();

            //ProvidedObjectID exported_psm_parameters = context.getProvidedObjectManager().createProvidedObject("Exported PSM Parameters", psmParameters, context);
            ProvidedObjectHelper.publish(context, "Exported PSM Parameters", psmParameters, PSMParameters.class, true);

            JOptionPane.showMessageDialog(null, "current parameters have been exported to the workspace");
        });

        add(exportButton);
    }

    private void setupApplySettings() {
        instantModeCheckBox = SlickerFactory.instance().createCheckBox("Instant Mode", settingsController.isInstantMode());
        instantModeCheckBox.setToolTipText("preemptively caches possible segments and applies settings immediately");
        instantModeCheckBox.addActionListener(e -> {
            settingsController.setInstantMode(instantModeCheckBox.isSelected());
        });

        lockSegmentOrderCheckBox = SlickerFactory.instance().createCheckBox("lock segment order", settingsController.getSegmentOrderSettings().usePredefined());
        lockSegmentOrderCheckBox.setToolTipText("lock current ordering of segments and only append new ones");
        lockSegmentOrderCheckBox.addActionListener(e -> {
            boolean selected = lockSegmentOrderCheckBox.isSelected();
            if (selected) {
                List<Segment> segmentOrder = psmUpdateManager.getSegmentOrderConnection().getSegmentOrder();
                if (segmentOrder == null)
                    segmentOrder = new ArrayList<>();
                SegmentOrderSettings segmentOrderSettings = new SegmentOrderSettings(segmentOrder);
                settingsController.setSegmentOrderSettings(segmentOrderSettings);
            } else
                settingsController.setSegmentOrderSettings(SegmentOrderSettings.HEURISTIC);
        });

        applySettingsButton = SlickerFactory.instance().createButton("apply settings");
        applySettingsButton.addActionListener(e -> settingsController.applySettings());

        unappliedChangesLabel = SlickerFactory.instance().createLabel("");
        unappliedChangesLabel.setIcon(checkmarkIcon);
        settingsController.getExecutionStatusConnection().register(this);

        JPanel applyPanel = SlickerFactory.instance().createRoundedPanel();
        applyPanel.add(instantModeCheckBox);
        applyPanel.add(lockSegmentOrderCheckBox);
        applyPanel.add(applySettingsButton);
        applyPanel.add(unappliedChangesLabel);
        add(applyPanel);
    }


    @Override
    public void update(ExecutionStatusConnection.ExecutionStatus status) {
        SwingUtilities.invokeLater(() -> {
            switch (status) {
                case Dirty:
                    unappliedChangesLabel.setIcon(gearIcon);
                    unappliedChangesLabel.setToolTipText("there are unapplied data changes not reflected in the visualization");
                    break;
                case Running:
                    unappliedChangesLabel.setIcon(loadingIcon);
                    unappliedChangesLabel.setToolTipText("settings are being applied");
                    break;
                case Done:
                    unappliedChangesLabel.setIcon(checkmarkIcon);
                    unappliedChangesLabel.setToolTipText("");
                    break;
            }
        });
    }


    private void setupFilterSettings() {
        JTextField filterSettingsText = new JTextField(10);
        filterSettingsText.setText(settingsController.getFilterSettings().toString());
        filterSettingsText.setEditable(false);

        JComboBox filterComboBox = SlickerFactory.instance().createComboBox(FilterType.values());
        filterComboBox.setSelectedItem(settingsController.getFilterSettings().getType());
        Runnable setFilterAction = () -> {
            FilterType type = (FilterType) filterComboBox.getSelectedItem();
            if (type == null) return;

            FilterSettings filterSettings = null;

            int userSelection = 0;

            switch (type) {
                case VariantPercentile:
                    DialogPanel<PercentageFilterSettings> panel1 = DialogPanel.createVariantPercentileFilter(settingsController.getFilterSettings(), settingsController.getExtraComputationsController());
                    userSelection = JOptionPane.showConfirmDialog(null, panel1, panel1.getTitle(), JOptionPane.OK_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE);
                    filterSettings = panel1.getResult();
                    break;
                case CaseDurationPercentile:
                    DialogPanel<PercentageFilterSettings> panel2 = DialogPanel.createDurationPercentileFilter(settingsController.getFilterSettings(), settingsController.getExtraComputationsController());
                    userSelection = JOptionPane.showConfirmDialog(null, panel2, panel2.getTitle(), JOptionPane.OK_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE);
                    filterSettings = panel2.getResult();
                    break;
                case CaseIds:
                    DialogPanel<CaseIdFilterSettings> panel3 = DialogPanel.createCaseIdFilter(settingsController.getFilterSettings());
                    userSelection = JOptionPane.showConfirmDialog(null, panel3, panel3.getTitle(), JOptionPane.OK_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE);
                    filterSettings = panel3.getResult();
                    break;
                case CaseDurationThreshold:
                    DialogPanel<DurationThresholdFilterSettings> panel4 = DialogPanel.createDurationThresholdFilter(settingsController.getFilterSettings(), settingsController.getExtraComputationsController());
                    userSelection = JOptionPane.showConfirmDialog(null, panel4, panel4.getTitle(), JOptionPane.OK_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE);
                    filterSettings = panel4.getResult();
                    break;
                case None:
                default:
                    filterSettings = FilterSettings.NONE;
                    break;
            }

            if (userSelection == 0 && filterSettings != null) {
                settingsController.setFilterSettings(filterSettings);
                filterSettingsText.setText(filterSettings.toString());
            } else if (filterSettings == null) {
                JOptionPane.showMessageDialog(null, "incorrect input format", "Incorrect Input Warning", JOptionPane.ERROR_MESSAGE);
            }
        };
        filterComboBox.addActionListener(e -> setFilterAction.run());
        filterSettingsText.addMouseListener(new MouseListener() {
            @Override
            public void mouseClicked(MouseEvent e) {
                setFilterAction.run();
            }

            @Override
            public void mousePressed(MouseEvent e) {

            }

            @Override
            public void mouseReleased(MouseEvent e) {

            }

            @Override
            public void mouseEntered(MouseEvent e) {

            }

            @Override
            public void mouseExited(MouseEvent e) {

            }
        });

        JPanel filterSettingsPanel = SlickerFactory.instance().createRoundedPanel();
        filterSettingsPanel.add(SlickerFactory.instance().createLabel("Filter"));
        filterSettingsPanel.add(filterComboBox);
        filterSettingsPanel.add(filterSettingsText);

        add(filterSettingsPanel);
    }


    private void setupAggregationSettings() {
        JCheckBox aggNormalizationCheckbox = SlickerFactory.instance().createCheckBox("normalize start time", settingsController.getAggregationIntervalSettings().isNormalizeProcessStart());
        aggNormalizationCheckbox.setToolTipText("normalizes process start time according to interval preset.\nE.g. first day of the year, quarter, month..");

        JComboBox aggIntervalType = SlickerFactory.instance().createComboBox(IntervalPresetType.values());
        aggIntervalType.setSelectedItem(settingsController.getAggregationIntervalSettings().getPresetType());
        Runnable aggregationIntervalAction = () -> {
            IntervalPresetType type = (IntervalPresetType) aggIntervalType.getSelectedItem();
            if (type == null) return;

            boolean normalize = aggNormalizationCheckbox.isSelected();

            SwingWorker<AggregationIntervalSettings, Void> sw = new SwingWorker<AggregationIntervalSettings, Void>() {
                @Override
                protected AggregationIntervalSettings doInBackground() throws Exception {
                    AggregationIntervalSettings aggSettings = null;
                    switch (type) {
                        case Custom:
                            CaseIdFilter filter = settingsController.getExtraComputationsController().getFilter(settingsController.getFilterSettings());
                            Tuple2<DateTime, DateTime> tuple2 = settingsController.getExtraComputationsController().getProcessDuration(filter);
                            if (tuple2._1 == null) return null;

                            String start = DateTimeFormat.shortDateTime().print(tuple2._1);
                            String end = DateTimeFormat.shortDateTime().print(tuple2._2);
                            Duration d = new Duration(tuple2._1, tuple2._2);
                            String duration = ToStrings.print(d);

                            String prevString = "2 weeks";
                            if (settingsController.getAggregationIntervalSettings().getPresetType() == IntervalPresetType.Custom) {
                                Period prev = settingsController.getAggregationIntervalSettings().getPeriod();
                                prevString = ToStrings.print(prev);
                            }

                            String s1 = JOptionPane.showInputDialog("Process starts from " + start + " to " + end + ".\nA total duration of " + duration + ".\nDuration of Intervals [x weeks, y days, etc.]", prevString);
                            if (s1 == null) return null;

                            try {
                                Period period = Period.parse(s1, PeriodFormat.getDefault());
                                aggSettings = new AggregationIntervalSettings(period);
                            } catch (IllegalArgumentException ex) {
                                JOptionPane.showMessageDialog(null, "Incorrect Format");
                            }
                            break;
                        default:
                            aggSettings = new AggregationIntervalSettings(type, normalize);
                            break;
                    }
                    return aggSettings;
                }

                @Override
                protected void done() {
                    try {
                        AggregationIntervalSettings aggSettings = get();
                        if (aggSettings != null) settingsController.setAggregationIntervalSettings(aggSettings);
                    } catch (InterruptedException | ExecutionException e) {
                        e.printStackTrace();
                    }
                }
            };

            sw.execute();

        };
        aggIntervalType.addActionListener(e -> {
            IntervalPresetType type = (IntervalPresetType) aggIntervalType.getSelectedItem();
            if (type == null) return;

            switch (type) {
                case Custom:
                    aggNormalizationCheckbox.setSelected(false);
                    aggNormalizationCheckbox.setEnabled(false);
                    break;
                default:
                    aggNormalizationCheckbox.setEnabled(true);
                    break;
            }

            aggregationIntervalAction.run();
        });
        aggNormalizationCheckbox.addActionListener(e -> aggregationIntervalAction.run());

        aggregationComboBox = SlickerFactory.instance().createComboBox(PSMAggregationSettings.AggregationVersion.values());
        aggregationComboBox.setSelectedItem(psmUpdateManager.getPsmAggregationSettings().getAggregationVersion());
        aggregationComboBox.addActionListener(e -> {
            PSMAggregationSettings.AggregationVersion selected = (PSMAggregationSettings.AggregationVersion) aggregationComboBox.getSelectedItem();
            if (selected != null) {
                psmUpdateManager.getPsmAggregationSettings().setAggregationVersion(selected);
                psmUpdateManager.visualSettingsChanged();
            }
        });

        JPanel aggregationIntervalPanel = SlickerFactory.instance().createRoundedPanel();
        aggregationIntervalPanel.add(SlickerFactory.instance().createLabel("Aggregation"));
        aggregationIntervalPanel.add(aggIntervalType);
        aggregationIntervalPanel.add(aggNormalizationCheckbox);
        aggregationIntervalPanel.add(SlickerFactory.instance().createLabel("Type"));
        aggregationIntervalPanel.add(aggregationComboBox);

        add(aggregationIntervalPanel);
    }


    private void setupClassificationSettings() {
        JTextField classificationSettingsText = new JTextField(10);
        classificationSettingsText.setText(settingsController.getClassificationSettings().toString());
        classificationSettingsText.setEditable(false);

        JComboBox classificationComboBox = SlickerFactory.instance().createComboBox(ClassificationType.values());
        classificationComboBox.setSelectedItem(settingsController.getClassificationSettings().getType());
        Runnable setClassificationAction = () -> {
            ClassificationType type = (ClassificationType) classificationComboBox.getSelectedItem();
            if (type == null) return;

            ClassificationSettings classificationSettings = null;

            int userSelection = 0;

            switch (type) {
                case VariantPercentile:
                    DialogPanel<PercentileClassificationSettings> panel1 = DialogPanel.createVariantPercentileClassification(settingsController.getClassificationSettings(), settingsController.getFilterSettings(), settingsController.getExtraComputationsController());
                    userSelection = JOptionPane.showConfirmDialog(null, panel1, panel1.getTitle(), JOptionPane.OK_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE);
                    classificationSettings = panel1.getResult();
                    break;
                case CaseDurationPercentile:
                    DialogPanel<PercentileClassificationSettings> panel2 = DialogPanel.createDurationPercentileClassification(settingsController.getClassificationSettings(), settingsController.getFilterSettings(), settingsController.getExtraComputationsController());
                    userSelection = JOptionPane.showConfirmDialog(null, panel2, panel2.getTitle(), JOptionPane.OK_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE);
                    classificationSettings = panel2.getResult();
                    break;
                case CaseDurationThreshold:
                    DialogPanel<DurationThresholdClassificationSettings> panel3 = DialogPanel.createThresholdGraphClassification(settingsController.getClassificationSettings(), settingsController.getFilterSettings(), settingsController.getExtraComputationsController());
                    userSelection = JOptionPane.showConfirmDialog(null, panel3, panel3.getTitle(), JOptionPane.OK_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE);
                    classificationSettings = panel3.getResult();
                    break;
                case CaseDurationClasses:
                    DialogPanel<DurationClassesClassificationSettings> panel4 = DialogPanel.createClassesGraphClassification(settingsController.getClassificationSettings(), settingsController.getFilterSettings(), settingsController.getExtraComputationsController());
                    userSelection = JOptionPane.showConfirmDialog(null, panel4, panel4.getTitle(), JOptionPane.OK_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE);
                    classificationSettings = panel4.getResult();
                    break;
                case None:
                default:
                    classificationSettings = ClassificationSettings.NONE;
                    break;
            }

            if (userSelection == 0 && classificationSettings != null) {
                settingsController.setClassificationSettings(classificationSettings);
                classificationSettingsText.setText(classificationSettings.toString());
            } else if (classificationSettings == null) {
                JOptionPane.showMessageDialog(null, "incorrect input format", "Incorrect Input Warning", JOptionPane.ERROR_MESSAGE);
            }
        };
        classificationComboBox.addActionListener(e -> setClassificationAction.run());
        classificationSettingsText.addMouseListener(new MouseListener() {
            @Override
            public void mouseClicked(MouseEvent e) {
                setClassificationAction.run();
            }

            @Override
            public void mousePressed(MouseEvent e) {

            }

            @Override
            public void mouseReleased(MouseEvent e) {

            }

            @Override
            public void mouseEntered(MouseEvent e) {

            }

            @Override
            public void mouseExited(MouseEvent e) {

            }
        });

        JButton legendButton = SlickerFactory.instance().createButton("legend");
        legendButton.addActionListener(e -> {
            SwingUtilities.invokeLater(() -> {
                legendButton.setEnabled(false);
                ClassificationLegendFrame frame = new ClassificationLegendFrame("Classification Window", psmUpdateManager, legendButton);
                frame.setVisible(true);
            });
        });

        JPanel classificationSettingsPanel = SlickerFactory.instance().createRoundedPanel();
        classificationSettingsPanel.add(SlickerFactory.instance().createLabel("Classification"));
        classificationSettingsPanel.add(classificationComboBox);
        classificationSettingsPanel.add(classificationSettingsText);
        classificationSettingsPanel.add(legendButton);

        add(classificationSettingsPanel);
    }

    public void clickApplySettings() {
        if (applySettingsButton != null)
            applySettingsButton.doClick();
    }

    public void setToInstantMode() {
        if (instantModeCheckBox != null)
            instantModeCheckBox.doClick();
    }

}
