package org.processmining.performancespectrumintegration.visualization;

import com.google.common.collect.HashBiMap;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.framework.util.ui.widgets.ProMSplitPane;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.elements.Arc;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.jgraph.visualization.ProMJGraphPanel;
import org.processmining.performancespectrumintegration.models.DataController;
import org.processmining.performancespectrumintegration.models.InteractivePSM;
import org.processmining.performancespectrumintegration.models.MyDataSource;
import org.processmining.performancespectrumintegration.models.settings.DataSettings;
import org.processmining.performancespectrumintegration.mypsm.PSMPanel;
import org.processmining.performancespectrumintegration.mypsm.update.PSMUpdateManager;
import org.processmining.performancespectrumintegration.util.Utils;
import org.processmining.performancespectrumintegration.visualization.controller.GraphController;
import org.processmining.performancespectrumintegration.visualization.controller.SettingsController;

import javax.swing.*;
import java.awt.*;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;

public class TopPanel extends JPanel implements ContentUpdateListener {

    private final Petrinet petrinet;
    private final PlaceSelectionMap placeSelectionMap;
    private final SettingsController settingsController;
    private final UIPluginContext context;
    private final PSMUpdateManager psmUpdateManager;
    private GraphController graphController;
    private PSMPanel psmPanel;

    private JFrame externalView;
    private JComponent visibleComponent;

    private boolean splitMode;

    public TopPanel(UIPluginContext context, InteractivePSM interactivePSM) {
        super(new BorderLayout());
        setPreferredSize(new Dimension(1920, 550));
        this.context = context;
        HashBiMap<Transition, Transition> transitionConnection = HashBiMap.create();
        HashBiMap<Place, Place> placesConnection = HashBiMap.create();
        HashBiMap<Arc, Arc> arcConnection = HashBiMap.create();
        this.petrinet = Utils.cloneNet(interactivePSM.getAcceptingNet().getNet(), placesConnection, transitionConnection, arcConnection); // maps are original -> clone
        PetrinetCopyConnection petrinetConnection = new PetrinetCopyConnection(transitionConnection, placesConnection, arcConnection);
        this.splitMode = true;
        DataController<MyDataSource> dataController = new DataController<>(interactivePSM, MyDataSource::new);
        this.psmUpdateManager = new PSMUpdateManager(dataController);
        this.placeSelectionMap = new PlaceSelectionMap(petrinet, psmUpdateManager, petrinetConnection);
        this.graphController = new GraphController(context, psmUpdateManager, placeSelectionMap, petrinet, petrinetConnection, splitMode);
        DataSettings initialDataSettings = DataSettings.defaultSettings();

        boolean initialInstantMode = false;
        if (interactivePSM.getPsmParameters() != null) {
            interactivePSM.getPsmParameters().recreate(initialDataSettings, placeSelectionMap, graphController.getSegmentMeasurementController(), interactivePSM.getAcceptingNet().getNet());
            initialInstantMode = interactivePSM.getPsmParameters().isInstantMode();
        }

        this.settingsController = new SettingsController(psmUpdateManager, dataController, initialDataSettings, placeSelectionMap, initialInstantMode);
        this.psmPanel = new PSMPanel(psmUpdateManager);

        graphController.setContentUpdateListener(this);
        graphController.initGraphVisualization();

        initView();
    }

    public void setSplitMode(boolean splitMode) {
        this.splitMode = splitMode;
        graphController.setSplitMode(splitMode);
        initView();
    }

    private void initView() {
        removeAll();

        if (splitMode) {
            if (externalView != null)
                externalView.dispose();

            ProMSplitPane proMSplitPane = new ProMSplitPane(ProMSplitPane.HORIZONTAL_SPLIT);
            ProMJGraphPanel graphPanel = graphController.getGraphPanel();
            graphPanel.setPreferredSize(new Dimension(750, 550));
            graphPanel.setMinimumSize(new Dimension(400, 550));
            proMSplitPane.setLeftComponent(graphPanel);
            proMSplitPane.setRightComponent(psmPanel);
            proMSplitPane.setDividerLocation(-1);
            proMSplitPane.revalidate();
            proMSplitPane.repaint();

            visibleComponent = proMSplitPane;
            add(visibleComponent);
        } else {
            visibleComponent = graphController.getGraphPanel();

            add(visibleComponent);

            externalView = new JFrame("External PSM");
            externalView.add(psmPanel);
            externalView.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
            //externalView.setSize(800, 600);
            externalView.setSize(psmPanel.getMinimumSize());
            externalView.addWindowListener(new WindowListener() {
                @Override
                public void windowOpened(WindowEvent e) {

                }

                @Override
                public void windowClosing(WindowEvent e) {

                }

                @Override
                public void windowClosed(WindowEvent e) {
                    SwingUtilities.invokeLater(() -> setSplitMode(true));
                }

                @Override
                public void windowIconified(WindowEvent e) {

                }

                @Override
                public void windowDeiconified(WindowEvent e) {

                }

                @Override
                public void windowActivated(WindowEvent e) {

                }

                @Override
                public void windowDeactivated(WindowEvent e) {

                }
            });
            externalView.pack();
            externalView.setVisible(true);
        }

        revalidate();
        repaint();
    }

    public void updateGraph() {
        if (splitMode && visibleComponent instanceof ProMSplitPane) {
            ((ProMSplitPane) visibleComponent).setLeftComponent(graphController.getGraphPanel());
        } else {
            removeAll();
            visibleComponent = graphController.getGraphPanel();
            add(visibleComponent);
        }
        visibleComponent.revalidate();
        visibleComponent.repaint();
        revalidate();
        repaint();
    }


    public SettingsController getSettingsController() {
        return settingsController;
    }

    @Override
    public void contentChanged(Object source) {
        if (source instanceof GraphController) {
            updateGraph();
        }
    }

    public void toggleSplitMode() {
        setSplitMode(!splitMode);
    }

    public PSMUpdateManager getPSMUpdateManager() {
        return psmUpdateManager;
    }
}
