package org.processmining.performancespectrumintegration.visualization.controller;

import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.models.connections.GraphLayoutConnection;
import org.processmining.models.graphbased.AttributeMap;
import org.processmining.models.graphbased.AttributeMapOwner;
import org.processmining.models.graphbased.ViewSpecificAttributeMap;
import org.processmining.models.graphbased.directed.DirectedGraphElement;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.jgraph.ProMJGraphVisualizer;
import org.processmining.models.jgraph.visualization.ProMJGraphPanel;
import org.processmining.performancespectrumintegration.models.settings.DataSettings;
import org.processmining.performancespectrumintegration.mypsm.update.PSMUpdateListener;
import org.processmining.performancespectrumintegration.mypsm.update.PSMUpdateManager;
import org.processmining.performancespectrumintegration.mypsm.update.PlaceSegmentConnection;
import org.processmining.performancespectrumintegration.visualization.ContentUpdateListener;
import org.processmining.performancespectrumintegration.visualization.PetrinetCopyConnection;
import org.processmining.performancespectrumintegration.visualization.PlaceSelectionMap;
import org.processmining.performancespectrumintegration.visualization.PlaceSelectionState;
import org.processmining.performancespectrumintegration.visualization.help.HelpPanels;

import javax.swing.*;
import java.awt.*;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.Collection;

public class GraphController implements PSMUpdateListener {

    private final UIPluginContext context;
    private final PSMUpdateManager psmUpdateManager;
    private final Petrinet petrinet;
    private final SegmentMeasurementController segmentMeasurementController;
    private final PlaceSelectionController placeSelectionController;
    private ViewSpecificAttributeMap map;
    private ProMJGraphPanel graphPanel;
    private boolean splitMode;
    private ContentUpdateListener listener;
    private Place lastHighlightedPlace;

    public GraphController(UIPluginContext context, PSMUpdateManager psmUpdateManager, PlaceSelectionMap placeSelectionMap, Petrinet petrinet, PetrinetCopyConnection petrinetCopyConnection, boolean splitMode) {
        this.context = context;
        this.psmUpdateManager = psmUpdateManager;
        this.petrinet = petrinet;
        this.placeSelectionController = new PlaceSelectionController(this, placeSelectionMap);
        this.segmentMeasurementController = new SegmentMeasurementController(this, placeSelectionMap, petrinet, petrinetCopyConnection);
        this.map = new ViewSpecificAttributeMap();
        this.splitMode = splitMode;

        psmUpdateManager.addListener(this);
    }

    public ProMJGraphPanel getGraphPanel() {
        return graphPanel;
    }

    public void setSplitMode(boolean splitMode) {
        this.splitMode = splitMode;
        initGraphVisualization();
    }

    private void initAttributeMap() {
        map.putViewSpecific(petrinet, AttributeMap.PREF_ORIENTATION, splitMode ? SwingConstants.NORTH : SwingConstants.WEST);
        for (Place place : petrinet.getPlaces()) {
            map.putViewSpecific(place, AttributeMap.MOVEABLE, false);
            map.putViewSpecific(place, AttributeMap.RESIZABLE, false);
        }
        for (Transition transition : petrinet.getTransitions()) {
            map.putViewSpecific(transition, AttributeMap.MOVEABLE, false);
            map.putViewSpecific(transition, AttributeMap.RESIZABLE, false);
        }
        placeSelectionController.updateAttributes(map);
        segmentMeasurementController.updateAttributes(map);
    }

    public void initGraphVisualization() {
        map = new ViewSpecificAttributeMap();
        initAttributeMap();

        //graphPanel = ProMJGraphVisualizer.instance().visualizeGraph(context, petrinet, map);
        graphPanel = ProMJGraphVisualizer.instance().visualizeGraphWithoutRememberingLayout(petrinet, map);

        //graphPanel.getViewInteractionPanels().values().remove(SwingConstants.SOUTH);

        graphPanel.setPreferredSize(new Dimension(splitMode ? 1920 / 2 : 1900, 550));
        graphPanel.addViewInteractionPanel(new HelpPanels.GraphLegendButtonPanel(), SwingConstants.SOUTH);
        graphPanel.addViewInteractionPanel(new HelpPanels.ControlsHelpButtonPanel(), SwingConstants.SOUTH);
        addListenersToGraph();
    }

    private void addListenersToGraph() {

        JPopupMenu measurementMenu = new JPopupMenu("segment measurement");
        JMenuItem resetSegments = new JMenuItem("reset all segments");
        resetSegments.addActionListener(e -> placeSelectionController.resetAll());
        JMenuItem resetSource = new JMenuItem("reset source set");
        resetSource.addActionListener(e -> segmentMeasurementController.clearSourceTransitions());
        JMenuItem resetTarget = new JMenuItem("reset target set");
        resetTarget.addActionListener(e -> segmentMeasurementController.clearTargetTransitions());
        JMenuItem measure = new JMenuItem("measure");
        measure.addActionListener(e -> {
            if (segmentMeasurementController.addMeasurementPlace(false, false) == null) {
                JOptionPane.showMessageDialog(graphPanel, "The source and target set you want to measure between need to be set first.");
            }
        });
        JMenuItem measure_with_hiding_subnet = new JMenuItem("measure with hiding subnet");
        measure_with_hiding_subnet.addActionListener(e -> {
            if (segmentMeasurementController.addMeasurementPlace(true, false) == null) {
                JOptionPane.showMessageDialog(graphPanel, "The source and target set you want to measure between need to be set first.\nHiding the subnet only works if it is on am isolated path between those selected sets.\nMeasurement places cannot be inside the subnet.");
            }
        });

        measurementMenu.add(resetSegments);
        measurementMenu.add(resetSource);
        measurementMenu.add(resetTarget);
        measurementMenu.add(measure);
        measurementMenu.add(measure_with_hiding_subnet);

        graphPanel.getGraph().setSelectionEnabled(false);

        graphPanel.getGraph().addMouseListener(new MouseListener() {

            private Place lastClicked;

            @Override
            public void mouseClicked(MouseEvent e) {
                /*
                if (SwingUtilities.isLeftMouseButton(e) && settingsController.isInstantMode()) {
                    DirectedGraphElement elementForLocation = graphPanel.getElementForLocation(e.getX(), e.getY());
                    if (elementForLocation instanceof Place) {
                        Place place = (Place) elementForLocation;
                        placeSelectionController.toggleState(place);
                    }

                }*/
                if (SwingUtilities.isLeftMouseButton(e)) {
                    PlaceSegmentConnection conn = psmUpdateManager.getPlaceSegmentConnection();
                    DirectedGraphElement elementForLocation = graphPanel.getElementForLocation(e.getX(), e.getY());

                    if (elementForLocation instanceof Place) {
                        conn.markPlace((Place) elementForLocation);
                        lastClicked = (Place) elementForLocation;
                    } else {
                        conn.markPlace(null);
                    }
                    conn.applyHighlighting();
                    psmUpdateManager.placeSegmentConnectionChanged();

                    graphPanel.getGraph().clearSelection();
                }
            }

            private void showPopup(MouseEvent e) {
                if (graphPanel.getElementForLocation(e.getX(), e.getY()) instanceof Place) {
                    Place place = (Place) graphPanel.getElementForLocation(e.getX(), e.getY());

                    JPopupMenu placeMenu = new JPopupMenu("state selection");
                    JMenuItem off = new JMenuItem("off");
                    off.addActionListener(e1 -> {
                        placeSelectionController.setState(place, PlaceSelectionState.Off);
                    });
                    JMenuItem individual = new JMenuItem("individual");
                    individual.addActionListener(e1 -> {
                        placeSelectionController.setState(place, PlaceSelectionState.Individual);
                    });
                    JMenuItem aggregated = new JMenuItem("aggregated");
                    aggregated.addActionListener(e1 -> {
                        placeSelectionController.setState(place, PlaceSelectionState.Aggregate);
                    });

                    placeMenu.add(off);
                    placeMenu.add(individual);
                    placeMenu.add(aggregated);

                    if (segmentMeasurementController.isArtificialPlace(place)) {
                        JMenuItem delete = new JMenuItem("delete");
                        delete.addActionListener(e1 -> {
                            segmentMeasurementController.removeMeasurementPlace(place);
                        });
                        placeMenu.add(delete);
                    }

                    placeMenu.show(e.getComponent(), e.getX(), e.getY());
                } else if (graphPanel.getElementForLocation(e.getX(), e.getY()) instanceof Transition) {
                    Transition transition = (Transition) graphPanel.getElementForLocation(e.getX(), e.getY());

                    JPopupMenu transitionMenu = new JPopupMenu("transition");
                    if (segmentMeasurementController.isSourceTransition(transition)) {
                        JMenuItem remove_from_source = new JMenuItem("remove from source set");
                        remove_from_source.addActionListener(e1 -> segmentMeasurementController.removeSourceTransition(transition));
                        transitionMenu.add(remove_from_source);

                    } else if (segmentMeasurementController.isTargetTransition(transition)) {
                        JMenuItem remove_from_target = new JMenuItem("remove from target set");
                        remove_from_target.addActionListener(e1 -> segmentMeasurementController.removeTargetTransition(transition));
                        transitionMenu.add(remove_from_target);

                    } else {
                        JMenuItem add_to_source = new JMenuItem("add to source set");
                        add_to_source.addActionListener(e1 -> segmentMeasurementController.addSourceTransition(transition));
                        JMenuItem add_to_target = new JMenuItem("add to target set");
                        add_to_target.addActionListener(e1 -> segmentMeasurementController.addTargetTransition(transition));
                        transitionMenu.add(add_to_source);
                        transitionMenu.add(add_to_target);
                    }

                    transitionMenu.show(e.getComponent(), e.getX(), e.getY());
                } else {
                    measurementMenu.show(e.getComponent(), e.getX(), e.getY());
                }
            }

            @Override
            public void mousePressed(MouseEvent e) {
                if (e.isPopupTrigger()) {
                    showPopup(e);
                }
            }

            @Override
            public void mouseReleased(MouseEvent e) {
                if (e.isPopupTrigger()) {
                    showPopup(e);
                }
            }

            @Override
            public void mouseEntered(MouseEvent e) {

            }

            @Override
            public void mouseExited(MouseEvent e) {

            }

        });


        graphPanel.getGraph().addKeyListener(new KeyListener() {
            @Override
            public void keyTyped(KeyEvent e) {

            }

            @Override
            public void keyPressed(KeyEvent e) {
                if (lastHighlightedPlace == null) return;
                switch (e.getKeyCode()) {
                    case KeyEvent.VK_DELETE:
                        segmentMeasurementController.removeMeasurementPlace(lastHighlightedPlace);
                        break;
                    case KeyEvent.VK_C:
                        placeSelectionController.toggleState(lastHighlightedPlace);
                        break;
                    default:
                        break;
                }
            }

            @Override
            public void keyReleased(KeyEvent e) {

            }
        });

    }

    public void updatedAttributes(AttributeMapOwner... owners) {
        SwingUtilities.invokeLater(() -> {
            if (graphPanel == null) return;
            GraphLayoutConnection layoutConnection = graphPanel.getGraph().getLayoutConnection();
            if (layoutConnection != null)
                layoutConnection.updatedAttributes(owners);
        });
    }

    public <T extends AttributeMapOwner> void updatedAttributes(Collection<T> owners) {
        updatedAttributes(owners.toArray(new AttributeMapOwner[0]));
    }

    public void updatedPetrinet() {
        SwingUtilities.invokeLater(() -> {
            initGraphVisualization();
            //graphPanel.getGraph().getGraphLayoutCache().reload();
            listener.contentChanged(this);
        });
    }

    public ViewSpecificAttributeMap getAttributeMap() {
        return map;
    }

    public void setContentUpdateListener(ContentUpdateListener listener) {
        this.listener = listener;
    }

    @Override
    public void onDataSourceChanged(DataSettings deltaSettings) {
        updateHighlighting();
    }

    @Override
    public void onPlaceSegmentConnectionChanged() {
        updateHighlighting();
    }

    private void updateHighlighting() {
        PlaceSegmentConnection conn = psmUpdateManager.getPlaceSegmentConnection();
        Place highlightedPlace = conn.getHighlightedPlace();
        if (lastHighlightedPlace != null && lastHighlightedPlace != highlightedPlace) {
            map.removeViewSpecific(lastHighlightedPlace, AttributeMap.STROKECOLOR);
            map.removeViewSpecific(lastHighlightedPlace, AttributeMap.BORDERWIDTH);
            updatedAttributes(lastHighlightedPlace);
            lastHighlightedPlace = null;
        }
        if (highlightedPlace != null && lastHighlightedPlace != highlightedPlace) {
            map.putViewSpecific(highlightedPlace, AttributeMap.STROKECOLOR, Color.red);
            map.putViewSpecific(highlightedPlace, AttributeMap.BORDERWIDTH, 3);
            updatedAttributes(highlightedPlace);
            lastHighlightedPlace = highlightedPlace;
        }

    }

    public SegmentMeasurementController getSegmentMeasurementController() {
        return segmentMeasurementController;
    }
}
