package org.processmining.performancespectrumintegration.visualization.controller;

import com.google.common.collect.BiMap;
import com.google.common.collect.HashBiMap;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import org.processmining.models.graphbased.AttributeMap;
import org.processmining.models.graphbased.AttributeMapOwner;
import org.processmining.models.graphbased.ViewSpecificAttributeMap;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Arc;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.shapes.Diamond;
import org.processmining.performancespectrumintegration.visualization.PetrinetCopyConnection;
import org.processmining.performancespectrumintegration.visualization.PlaceSelectionMap;

import java.awt.*;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

public class SegmentMeasurementController {

    private final GraphController graphController;
    private final PlaceSelectionMap placeSelectionMap;
    private final Map<Place, PetrinetCopyConnection> artificialPlaceSubnetMap;
    private final Petrinet petrinet;
    private final PetrinetCopyConnection petrinetConnection;
    private Set<Transition> sourceTransitions, targetTransitions;

    public SegmentMeasurementController(GraphController graphController, PlaceSelectionMap placeSelectionMap, Petrinet petrinet, PetrinetCopyConnection petrinetConnection) {
        this.graphController = graphController;
        this.placeSelectionMap = placeSelectionMap;
        this.petrinet = petrinet;
        this.petrinetConnection = petrinetConnection;
        artificialPlaceSubnetMap = Maps.newHashMap();
    }

    public void clearTargetTransitions() {
        if (targetTransitions != null) {
            for (Transition transition : targetTransitions) {
                graphController.getAttributeMap().removeViewSpecific(transition, AttributeMap.FILLCOLOR);
            }
            graphController.updatedAttributes(targetTransitions);
            targetTransitions.clear();
        }
    }

    public void clearSourceTransitions() {
        if (sourceTransitions != null) {
            for (Transition transition : sourceTransitions) {
                graphController.getAttributeMap().removeViewSpecific(transition, AttributeMap.FILLCOLOR);
            }
            graphController.updatedAttributes(sourceTransitions);
            sourceTransitions.clear();
        }
    }

    public boolean setTargetTransitions(Set<Transition> targetTransitions) {
        if (targetTransitions == null || targetTransitions.isEmpty()) return false;
        clearTargetTransitions();
        this.targetTransitions = targetTransitions;
        for (Transition transition : targetTransitions) {
            graphController.getAttributeMap().putViewSpecific(transition, AttributeMap.FILLCOLOR, Color.pink);
        }
        graphController.updatedAttributes(targetTransitions);
        return true;
    }

    public void addTargetTransition(Transition transition) {
        removeSourceTransition(transition);
        if (targetTransitions == null) targetTransitions = Sets.newHashSet();
        targetTransitions.add(transition);
        graphController.getAttributeMap().putViewSpecific(transition, AttributeMap.FILLCOLOR, Color.pink);
        graphController.updatedAttributes(transition);
    }

    public boolean setSourceTransitions(Set<Transition> sourceTransitions) {
        if (sourceTransitions == null || sourceTransitions.isEmpty()) return false;
        clearSourceTransitions();
        this.sourceTransitions = sourceTransitions;
        for (Transition transition : sourceTransitions) {
            graphController.getAttributeMap().putViewSpecific(transition, AttributeMap.FILLCOLOR, Color.orange);
        }
        graphController.updatedAttributes(sourceTransitions);
        return true;
    }

    public void addSourceTransition(Transition transition) {
        removeTargetTransition(transition);
        if (sourceTransitions == null) sourceTransitions = Sets.newHashSet();
        sourceTransitions.add(transition);
        graphController.getAttributeMap().putViewSpecific(transition, AttributeMap.FILLCOLOR, Color.orange);
        graphController.updatedAttributes(transition);
    }

    public boolean isArtificialPlace(Place place) {
        return artificialPlaceSubnetMap.containsKey(place);
    }

    public void removeSourceTransition(Transition transition) {
        if (sourceTransitions != null) {
            sourceTransitions.remove(transition);
            graphController.getAttributeMap().removeViewSpecific(transition, AttributeMap.FILLCOLOR);
            graphController.updatedAttributes(transition);
        }
    }

    public void removeTargetTransition(Transition transition) {
        if (targetTransitions != null) {
            targetTransitions.remove(transition);
            graphController.getAttributeMap().removeViewSpecific(transition, AttributeMap.FILLCOLOR);
            graphController.updatedAttributes(transition);
        }
    }

    public boolean removeMeasurementPlace(Place place) {
        if (artificialPlaceSubnetMap.containsKey(place)) {
            placeSelectionMap.removeState(place, graphController.getAttributeMap());

            for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> inEdge : petrinet.getInEdges(place)) {
                petrinet.removeArc(inEdge.getSource(), inEdge.getTarget());
            }
            for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> outEdge : petrinet.getOutEdges(place)) {
                petrinet.removeArc(outEdge.getSource(), outEdge.getTarget());
            }
            petrinet.removePlace(place);
            //graphController.getGraphPanel().getGraph().getModel().remove(new Object[]{place});

            // re-add removed subnet
            readdHiddenSubnet(place);

            graphController.getAttributeMap().clearViewSpecific(place);
            System.out.println("Removed place " + place);
            graphController.updatedPetrinet();

            return true;
        } else return false;
    }

    private void readdHiddenSubnet(Place place) {
        PetrinetCopyConnection between = artificialPlaceSubnetMap.remove(place);

        if (between == null) return;

        for (Place p : between.getOriginalPlaces()) {
            Place newCopy = petrinet.addPlace(p.getLabel());
            Place oldCopy = petrinetConnection.getCopyPlace(p);
            //copyAttributeMap(oldCopy, newCopy);
            removeAttributeMap(oldCopy);
            placeSelectionMap.addState(newCopy, graphController.getAttributeMap());
            petrinetConnection.putCopyPlace(p, newCopy);
        }
        for (Transition t : between.getOriginalTransitions()) {
            Transition newCopy = petrinet.addTransition(t.getLabel());
            newCopy.setInvisible(t.isInvisible());
            Transition oldCopy = petrinetConnection.getCopyTransition(t);
            //copyAttributeMap(oldCopy, newCopy);
            removeAttributeMap(oldCopy);
            petrinetConnection.putCopyTransition(t, newCopy);
        }
        for (Arc a : between.getOriginalArcs()) {
            Arc newCopy;
            if (a.getSource() instanceof Place) {
                Place source = (Place) a.getSource();
                Transition target = (Transition) a.getTarget();
                Place copySource = petrinetConnection.getCopyPlace(source);
                Transition copyTarget = petrinetConnection.getCopyTransition(target);
                newCopy = petrinet.addArc(copySource, copyTarget);
            } else {
                Transition source = (Transition) a.getSource();
                Place target = (Place) a.getTarget();
                Transition copySource = petrinetConnection.getCopyTransition(source);
                Place copyTarget = petrinetConnection.getCopyPlace(target);
                newCopy = petrinet.addArc(copySource, copyTarget);
            }
            Arc oldCopy = petrinetConnection.getCopyArc(a);
            //copyAttributeMap(oldCopy, newCopy);
            removeAttributeMap(oldCopy);
            petrinetConnection.putCopyArc(a, newCopy);
        }
    }

    private void copyAttributeMap(AttributeMapOwner original, AttributeMapOwner copy) {
        for (String key : original.getAttributeMap().keySet()) {
            copy.getAttributeMap().put(key, original.getAttributeMap().get(key));
        }
        ViewSpecificAttributeMap map = graphController.getAttributeMap();
        for (String key : map.getMapFor(original).keySet()) {
            map.putViewSpecific(copy, key, map.get(original, key));
        }
    }

    private void removeAttributeMap(AttributeMapOwner owner) {
        graphController.getAttributeMap().clearViewSpecific(owner);
    }

    public Place addMeasurementPlace(boolean removeSubnet, boolean silent) {
        Place createdPlace = null;
        if (checkSelection()) {

            PetrinetCopyConnection between = null;
            if (removeSubnet) {
                between = crawlBetween(sourceTransitions, targetTransitions);
                if (between == null)
                    return null;
                else
                    removeSubnet(between);
            }

            createdPlace = addPlace();

            artificialPlaceSubnetMap.put(createdPlace, between);
        }

        if (createdPlace != null) {
            System.out.println("Added place between " + sourceTransitions + " to " + targetTransitions);

            clearSourceTransitions();
            clearTargetTransitions();

            if (!silent)
                graphController.updatedPetrinet();
        }

        return createdPlace;
    }

    public PetrinetCopyConnection crawlBetween(final Set<Transition> source, final Set<Transition> target) {
        BiMap<Transition, Transition> crawledTransitions = HashBiMap.create(); // original -> copy
        BiMap<Place, Place> crawledPlaces = HashBiMap.create();
        BiMap<Arc, Arc> crawledArcs = HashBiMap.create();

        Set<Transition> frontier = source;
        int counter = 0;
        while (!frontier.isEmpty()) {
            if (counter >= 100) {
                System.out.println("Search in between transitions had to be cancelled due to max iterations being reached");
                return null;
            }

            System.out.println("Iteration: " + counter);

            Set<Place> places = transitionsPostSet(frontier, crawledArcs);
            if (places == null) {
                System.out.println("measurement place cannot be inside hidden subnet");
                return null;
            }
            places.removeAll(crawledPlaces.values());

            places.forEach(p -> {
                crawledPlaces.put(petrinetConnection.getOriginalPlace(p), p);
            });
            System.out.println("Intermediate Places: " + places);

            Set<Transition> transitions = placesPostSet(places, crawledArcs);
            if (transitions == null) {
                System.out.println("measurement place cannot be inside hidden subnet");
                return null;
            }

            // we looped back to the start transitions
            if (transitions.stream().anyMatch(source::contains)) {
                System.out.println("looped back around to start transitions");
                return null;
            }

            transitions.removeAll(target);
            transitions.removeAll(crawledTransitions.values());

            transitions.forEach(t -> {
                crawledTransitions.put(petrinetConnection.getOriginalTransition(t), t);
            });
            System.out.println("Intermediate Transitions: " + transitions);

            frontier = transitions;
            counter++;

            System.out.println("Current frontier: " + frontier);
        }
        if (checkPreSetCompleteness(source, crawledTransitions.values(), crawledPlaces.values())) {
            return new PetrinetCopyConnection(crawledTransitions, crawledPlaces, crawledArcs);
        } else {
            System.out.println("preset was not complete");
            return null;
        }
    }

    private boolean checkPreSetCompleteness(Set<Transition> source, Set<Transition> transitions, Set<Place> places) {
        for (Transition t : transitions) {
            if (!petrinet.getInEdges(t).stream().map(PetrinetEdge::getSource).map(Place.class::cast).allMatch(places::contains)) {
                System.out.println("preset of " + t + " is not complete");
                return false;
            }
        }
        for (Place p : places) {
            if (!petrinet.getInEdges(p).stream().map(PetrinetEdge::getSource).map(Transition.class::cast).allMatch(t -> transitions.contains(t) || source.contains(t))) {
                System.out.println("preset of " + p + " is not complete");
                return false;
            }
        }
        return true;
    }

    private Set<Place> transitionsPostSet(Set<Transition> transitions, BiMap<Arc, Arc> crawledArcs) {
        Set<Place> set = new HashSet<>();
        for (Transition t : transitions) {
            for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> petrinetEdge : petrinet.getOutEdges(t)) {
                Place place = (Place) petrinetEdge.getTarget();
                Arc arc = petrinet.getArc(t, place);
                if (!crawledArcs.containsValue(arc)) {
                    if (petrinetConnection.hasOriginalPlace(place) && petrinetConnection.hasOriginalArc(arc)) {
                        crawledArcs.put(petrinetConnection.getOriginalArc(arc), arc);
                        set.add(place);
                    } else return null; // art. place
                }
            }
        }
        return set;
    }

    private Set<Transition> placesPostSet(Set<Place> places, BiMap<Arc, Arc> crawledArcs) {
        Set<Transition> set = new HashSet<>();
        for (Place p : places) {
            for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> petrinetEdge : petrinet.getOutEdges(p)) {
                Transition transition = (Transition) petrinetEdge.getTarget();
                Arc arc = petrinet.getArc(p, transition);
                if (!crawledArcs.containsValue(arc)) {
                    if (petrinetConnection.hasOriginalTransition(transition) && petrinetConnection.hasOriginalArc(arc)) {
                        crawledArcs.put(petrinetConnection.getOriginalArc(arc), arc);
                        set.add(transition);
                    } else return null; // art. place
                }
            }
        }
        return set;
    }

    private void removeSubnet(PetrinetCopyConnection between) {
        between.getCopyArcs().forEach(arc -> petrinet.removeArc(arc.getSource(), arc.getTarget()));
        between.getCopyTransitions().forEach(petrinet::removeTransition);
        between.getCopyPlaces().forEach(petrinet::removePlace);
        placeSelectionMap.removeStates(between.getCopyPlaces(), graphController.getAttributeMap());
    }

    private boolean checkSelection() {
        return sourceTransitions != null && targetTransitions != null && !sourceTransitions.isEmpty() && !targetTransitions.isEmpty();
    }

    private Place addPlace() {
        if (checkSelection()) {
            Place createdPlace = petrinet.addPlace("measurement place");
            try {
                sourceTransitions.forEach(t -> petrinet.addArc(t, createdPlace));
                targetTransitions.forEach(t -> petrinet.addArc(createdPlace, t));
            } catch (Exception e) { // IllegalArgument may be thrown
                e.printStackTrace();
            }

            placeSelectionMap.addState(createdPlace, graphController.getAttributeMap());

            updateAttributes(createdPlace, graphController.getAttributeMap());

            return createdPlace;
        } else
            return null;
    }

    public void updateAttributes(ViewSpecificAttributeMap map) {
        for (Place place : artificialPlaceSubnetMap.keySet()) {
            updateAttributes(place, map);
        }
    }

    private void updateAttributes(Place place, ViewSpecificAttributeMap map) {
        map.putViewSpecific(place, AttributeMap.SHAPE, new Diamond());
    }

    public boolean isSourceTransition(Transition transition) {
        return sourceTransitions != null && sourceTransitions.contains(transition);
    }

    public boolean isTargetTransition(Transition transition) {
        return targetTransitions != null && targetTransitions.contains(transition);
    }

}
