package org.processmining.performancespectrumintegration.visualization.controller;

import org.joda.time.Duration;
import org.joda.time.Instant;
import org.processmining.performancespectrumintegration.algorithms.segments.Segment;
import org.processmining.performancespectrumintegration.models.DataController;
import org.processmining.performancespectrumintegration.models.MyDataSource;
import org.processmining.performancespectrumintegration.models.settings.*;
import org.processmining.performancespectrumintegration.mypsm.update.PSMUpdateManager;
import org.processmining.performancespectrumintegration.parameters.PSMParameters;
import org.processmining.performancespectrumintegration.visualization.PlaceSelectionMap;

import javax.swing.*;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import java.util.Set;
import java.util.concurrent.ExecutionException;

import static org.processmining.performancespectrumintegration.visualization.controller.ExecutionStatusConnection.ExecutionStatus.*;

public class SettingsController implements ChangeListener {

    private final DataController<MyDataSource> dataController;
    private final DataSettings dataSettings;
    private final PSMUpdateManager psmUpdateManager;
    private final PlaceSelectionMap placeSelectionMap;
    private final ExecutionStatusConnection executionStatusConnection;
    private DataSettings deltaDataSettings;
    private boolean firstRun;
    private boolean instantMode;
    private SwingWorker<Void, Void> currentWorker;

    public SettingsController(PSMUpdateManager psmUpdateManager, DataController<MyDataSource> dataController, DataSettings initialDataSettings, PlaceSelectionMap placeSelectionMap, boolean instantMode) {
        this.dataController = dataController;
        this.psmUpdateManager = psmUpdateManager;
        this.placeSelectionMap = placeSelectionMap;
        placeSelectionMap.setChangeListener(this);
        dataSettings = initialDataSettings;
        deltaDataSettings = new DataSettings();
        firstRun = true;
        this.instantMode = instantMode;
        executionStatusConnection = new ExecutionStatusConnection();
    }

    // place selection changed
    public void stateChanged(ChangeEvent e) {
        settingChanged();
    }

    public boolean isInstantMode() {
        return instantMode;
    }

    public void setInstantMode(boolean instantMode) {
        this.instantMode = instantMode;
        if (instantMode) {
            SwingWorker<Void, Void> sw = new SwingWorker<Void, Void>() {
                @Override
                protected Void doInBackground() throws Exception {
                    dataController.prepareCache(placeSelectionMap.computeAllPossibleSingleSegments());
                    return null;
                }
            };
            sw.execute();
        }
    }

    private void settingChanged() {
        executionStatusConnection.setStatus(Dirty);
        if (instantMode) applySettingsExploratively();
    }

    public Set<Segment> getSelectedSegments() {
        return deltaDataSettings.getSelectedSegments() != null ? deltaDataSettings.getSelectedSegments() : dataSettings.getSelectedSegments();
    }

    public void setSelectedSegments(Set<Segment> selectedSegments) {
        synchronized (dataSettings) {
            if (selectedSegments.equals(dataSettings.getSelectedSegments())) {
                deltaDataSettings.setSelectedSegments(null);
            } else {
                deltaDataSettings.setSelectedSegments(selectedSegments);
            }
        }
    }

    public FilterSettings getFilterSettings() {
        return deltaDataSettings.getFilterSettings() != null ? deltaDataSettings.getFilterSettings() : dataSettings.getFilterSettings();
    }

    public void setFilterSettings(FilterSettings filterSettings) {
        synchronized (dataSettings) {
            if (filterSettings.equals(dataSettings.getFilterSettings())) {
                deltaDataSettings.setFilterSettings(null);
            } else {
                deltaDataSettings.setFilterSettings(filterSettings);
                settingChanged();
            }
        }
    }

    public ClassificationSettings getClassificationSettings() {
        return deltaDataSettings.getClassificationSettings() != null ? deltaDataSettings.getClassificationSettings() : dataSettings.getClassificationSettings();
    }

    public void setClassificationSettings(ClassificationSettings classificationSettings) {
        synchronized (dataSettings) {
            if (classificationSettings.equals(dataSettings.getClassificationSettings())) {
                deltaDataSettings.setClassificationSettings(null);
            } else {
                deltaDataSettings.setClassificationSettings(classificationSettings);
                settingChanged();
            }
        }
    }

    public SegmentOrderSettings getSegmentOrderSettings() {
        return deltaDataSettings.getSegmentOrderSettings() != null ? deltaDataSettings.getSegmentOrderSettings() : dataSettings.getSegmentOrderSettings();
    }

    public void setSegmentOrderSettings(SegmentOrderSettings segmentOrderSettings) {
        synchronized (dataSettings) {
            if (segmentOrderSettings.equals(dataSettings.getSegmentOrderSettings())) {
                deltaDataSettings.setSegmentOrderSettings(null);
            } else {
                deltaDataSettings.setSegmentOrderSettings(segmentOrderSettings);
                settingChanged();
            }
        }
    }

    public PSMParameters getPSMParameters() {
        return new PSMParameters(dataSettings, placeSelectionMap, instantMode);
    }

    public void applySettings() {
        synchronized (dataSettings) {
            if (currentWorker != null && !currentWorker.isDone()) return;

            Set<Segment> segments = placeSelectionMap.computeSegments();

            setSelectedSegments(segments);

            dataSettings.apply(deltaDataSettings);

            DataSettings settingsDelta;
            if (firstRun) {
                settingsDelta = dataSettings.copy();
                firstRun = false;
            } else settingsDelta = deltaDataSettings.copy();

            deltaDataSettings = new DataSettings();

            executionStatusConnection.setStatus(Running);

            psmUpdateManager.dataSourceGoingToChange(settingsDelta);

            currentWorker = new SwingWorker<Void, Void>() {
                @Override
                protected Void doInBackground() throws Exception {
                    System.out.println("Applying Miner with " + settingsDelta);
                    Instant t1 = Instant.now();
                    try {
                        dataController.applyNewSettingsSequentially(settingsDelta);
                    } catch (Exception e) {
                        e.printStackTrace();
                    }
                    System.out.println("Mined new Data in " + new Duration(t1, Instant.now()));
                    return null;
                }

                @Override
                protected void done() {

                    try {
                        get();

                        executionStatusConnection.setStatus(Done);

                        psmUpdateManager.dataSourceChanged(settingsDelta);

                    } catch (InterruptedException | ExecutionException e) {
                        e.printStackTrace();
                    }

                }
            };

            currentWorker.execute();
        }
    }

    public void applySettingsExploratively() {
        //if(deltaDataSettings.isOnlySegmentOrderingSet() && deltaDataSettings.getSegmentOrderSettings().usePredefined()) return; // TODO if only the current ordering is locked, no need to update
        applySettings();
    }

    public AggregationIntervalSettings getAggregationIntervalSettings() {
        return deltaDataSettings.getAggregationIntervalSettings() != null ? deltaDataSettings.getAggregationIntervalSettings() :
                dataSettings.getAggregationIntervalSettings();
    }

    public void setAggregationIntervalSettings(AggregationIntervalSettings aggregationIntervalSettings) {
        synchronized (dataSettings) {
            if (aggregationIntervalSettings.equals(dataSettings.getAggregationIntervalSettings())) {
                deltaDataSettings.setAggregationIntervalSettings(null);
            } else {
                deltaDataSettings.setAggregationIntervalSettings(aggregationIntervalSettings);
                settingChanged();
            }
        }

    }

    public ExtraComputationsController getExtraComputationsController() {
        return dataController.getExtraComputationsController();
    }

    public ExecutionStatusConnection getExecutionStatusConnection() {
        return executionStatusConnection;
    }

}



